import numpy as np
import matplotlib.pyplot as plt

import xtrack as xt
import xobjects as xo
import xpart as xp

import RF_Track as RFT

#############################################
##########    Input parameters    ###########
#############################################

# Bunch parameters
mass = xp.ELECTRON_MASS_EV # eV/c^2
q0 = -1 # electrons
P0c = 50e6 # reference momentum, eV/c
P_over_q = P0c / q0 # V/c, reference rigidity

# FODO cell paramters
mu = np.pi/2 # rad, phase advance
Lcell = 2 # m, fodo cell length
Lquad = 0 # m, a thin quad

Ldrift = Lcell/2 - Lquad # m
k1L = np.sin(mu/2) / (Lcell/4) # 1/m, analytic quad integrated strength

#############################################
#######  RF-Track's part starts here  #######
#############################################

# Setup the elements
strength = k1L * P_over_q / 1e6 # MeV/m
Qf = RFT.Quadrupole(Lquad/2, strength/2) # 1/2 quad
QD = RFT.Quadrupole(Lquad, -strength) # full quad
Dr = RFT.Drift(Ldrift)
Dr.set_tt_nsteps(100) # for plots

# Setup the lattice
FODO = RFT.Lattice()
FODO.append(Qf) # 1/2 quad
FODO.append(Dr)
FODO.append(QD) # full quad
FODO.append(Dr)
FODO.append(Qf) # 1/2 quad

#############################################
#######  RF-Track's part ends here    #######
#############################################

# Back to Xsuite

elements = {
    'fodo': xt.RFT_Element(element=FODO),
}

# Build the line
line = xt.Line(elements=elements, element_names=list(elements.keys()))
line.particle_ref = xt.Particles(p0c=P0c, mass0=xt.ELECTRON_MASS_EV, q0=q0)

# Define the beam using the Twiss parameters
beta_gamma = P0c / mass
norm_emitt_x = 0.001 # mm.mrad, normalized emittance
norm_emitt_y = 0.001 # mm.mrad
geom_emitt_x = norm_emitt_x / beta_gamma / 1e6 # m.rad, geometric emittance
geom_emitt_y = norm_emitt_x / beta_gamma / 1e6 # m.rad

beta_x = Lcell * (1 + np.sin(mu/2)) / np.sin(mu) # m, analytic matched beta
beta_y = Lcell * (1 - np.sin(mu/2)) / np.sin(mu) # m

## Build particle object on context
n_part = 10000

rng = np.random.default_rng(12345)

# track
## Choose a context
context = xo.ContextCpu()         # For CPU
# context = xo.ContextCupy()      # For CUDA GPUs
# context = xo.ContextPyopencl()  # For OpenCL GPUs

## Transfer lattice on context and compile tracking code
line.build_tracker(_context=context)

particles = xp.Particles(
    _context=context,
    x=rng.normal(scale=np.sqrt(geom_emitt_x*beta_x), size=n_part),
    y=rng.normal(scale=np.sqrt(geom_emitt_y*beta_y), size=n_part),
    px=rng.normal(scale=np.sqrt(geom_emitt_x/beta_x), size=n_part),
    py=rng.normal(scale=np.sqrt(geom_emitt_y/beta_y), size=n_part),
    p0c=P0c, mass0=xt.ELECTRON_MASS_EV, q0=q0)

#############################################
#########   Xsuite tracking    ##############
#############################################

print('tracking starts')
line.track(particles)
print('tracking ends')

#############################################
## Retrieve Twiss from RF-Track and plot  ###
#############################################

# Retrieve the Twiss plot and the phase space
T = FODO.get_transport_table('%S %beta_x %beta_y')
plt.figure(1)
plt.plot(T[:,0], T[:,1], 'b-', linewidth=2, label='beta x')
plt.plot(T[:,0], T[:,2], 'r-', linewidth=2, label='beta y')
plt.legend(loc="upper left")
plt.xlabel('S [m]')
plt.ylabel('beta [m]')
plt.show()






# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt
import numpy as np

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()

collider.lhcb1.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['reverse'] = True

line = collider.lhcb2 # <- use lhcb2 to test the reverse option

tw8_closed = line.twiss(start='ip8')
tw8_open = line.twiss(start='ip8', betx=1.5, bety=1.5)

tw_part = line.twiss(start='ip8', end='ip2', zero_at='ip1', init='full_periodic')

# Test get strengths (to be moved to another script)
line.get_strengths().rows['mbw\..*l3.b2'].cols['k0l angle_rad']
# is:
# Table: 6 rows, 3 cols
# name                            k0l    angle_rad
# mbw.f6l3.b2            -0.000188729 -0.000188729
# mbw.e6l3.b2            -0.000188729 -0.000188729
# mbw.d6l3.b2            -0.000188729 -0.000188729
# mbw.c6l3.b2             0.000188729  0.000188729
# mbw.b6l3.b2             0.000188729  0.000188729
# mbw.a6l3.b2             0.000188729  0.000188729

line.get_strengths(reverse=False).rows['mbw\..*l3.b2'].cols['k0l angle_rad']
# is:
# Table: 6 rows, 3 cols
# name                            k0l    angle_rad
# mbw.a6l3.b2            -0.000188729 -0.000188729
# mbw.b6l3.b2            -0.000188729 -0.000188729
# mbw.c6l3.b2            -0.000188729 -0.000188729
# mbw.d6l3.b2             0.000188729  0.000188729
# mbw.e6l3.b2             0.000188729  0.000188729
# mbw.f6l3.b2             0.000188729  0.000188729

import xobjects as xo
str_table_rev = line.get_strengths() # Takes reverse from twiss_default
xo.assert_allclose(line['mbw.a6l3.b2'].k0,
        -str_table_rev['k0l', 'mbw.a6l3.b2'] / str_table_rev['length', 'mbw.a6l3.b2'],
        rtol=0, atol=1e-14)
xo.assert_allclose(line['mbw.a6l3.b2'].h,
        -str_table_rev['angle_rad', 'mbw.a6l3.b2'] / str_table_rev['length', 'mbw.a6l3.b2'],
        rtol=0, atol=1e-14)

str_table = line.get_strengths(reverse=False) # Takes reverse from twiss_default
xo.assert_allclose(line['mbw.a6l3.b2'].k0,
        str_table['k0l', 'mbw.a6l3.b2'] / str_table['length', 'mbw.a6l3.b2'],
        rtol=0, atol=1e-14)
xo.assert_allclose(line['mbw.a6l3.b2'].h,
        str_table['angle_rad', 'mbw.a6l3.b2'] / str_table['length', 'mbw.a6l3.b2'],
        rtol=0, atol=1e-14)

tw = line.twiss()

for tw8 in [tw8_closed, tw8_open]:
    assert tw8.name[-1] == '_end_point'
    assert np.all(tw8.rows['ip.?'].name
        == np.array(['ip8', 'ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7']))

    for nn in ['s', 'mux', 'muy']:
        assert np.all(np.diff(tw8.rows['ip.?'][nn]) > 0)
        assert tw8[nn][0] == 0
        xo.assert_allclose(tw8[nn][-1], tw[nn][-1], rtol=1e-12, atol=5e-7)

    xo.assert_allclose(
            tw8['betx', ['ip8', 'ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7']],
            tw[ 'betx', ['ip8', 'ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7']],
            rtol=1e-5, atol=0)

tw_part1 = line.twiss(start='ip8', end='ip2', zero_at='ip1', init='full_periodic')

assert tw_part1.name[0] == 'ip8'
assert tw_part1.name[-2] == 'ip2'
assert tw_part1.name[-1] == '_end_point'

for kk in ['s', 'mux', 'muy']:
    tw_part1[kk, 'ip1'] == 0.
    assert np.all(np.diff(tw_part1[kk]) >= 0)
    xo.assert_allclose(
        tw_part1[kk, 'ip8'], -(tw[kk, '_end_point'] - tw[kk, 'ip8']),
        rtol=1e-12, atol=5e-7)
    xo.assert_allclose(
        tw_part1[kk, 'ip2'], tw[kk, 'ip2'] - tw[kk, 0],
        rtol=1e-12, atol=5e-7)

tw_part2 = line.twiss(start='ip8', end='ip2', init='full_periodic')

assert tw_part2.name[0] == 'ip8'
assert tw_part2.name[-2] == 'ip2'
assert tw_part2.name[-1] == '_end_point'

for kk in ['s', 'mux', 'muy']:
    tw_part2[kk, 'ip8'] == 0.
    assert np.all(np.diff(tw_part2[kk]) >= 0)
    xo.assert_allclose(
        tw_part2[kk, 'ip2'],
        tw[kk, 'ip2'] - tw[kk, 'ip1'] +(tw[kk, '_end_point'] - tw[kk, 'ip8']),
        rtol=1e-12, atol=5e-7)



import json
import pathlib

import numpy as np
import pandas as pd
from cpymad.madx import Madx

import xdeps as xd
import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts
from xtrack.slicing import Teapot, Strategy

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_psb_chicane(test_context):
    mad = Madx(stdout=False)

    # Load mad model and apply element shifts
    mad.input(f'''
    call, file = '{str(test_data_folder)}/psb_chicane/psb.seq';
    call, file = '{str(test_data_folder)}/psb_chicane/psb_fb_lhc.str';

    beam, particle=PROTON, pc=0.5708301551893517;
    use, sequence=psb1;

    select,flag=error,clear;
    select,flag=error,pattern=bi1.bsw1l1.1*;
    ealign, dx=-0.0057;

    select,flag=error,clear;
    select,flag=error,pattern=bi1.bsw1l1.2*;
    select,flag=error,pattern=bi1.bsw1l1.3*;
    select,flag=error,pattern=bi1.bsw1l1.4*;
    ealign, dx=-0.0442;

    twiss;
    ''')

    line = xt.Line.from_madx_sequence(mad.sequence.psb1,
                                allow_thick=True,
                                enable_align_errors=True,
                                deferred_expressions=True)
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV,
                                gamma0=mad.sequence.psb1.beam.gamma)
    line.twiss_default['method'] = '4d'
    line.configure_bend_model(core='adaptive')

    # Build chicane knob (k0)
    line.vars['bsw_k0l'] = 0
    line.vars['k0bi1bsw1l11'] = (line.vars['bsw_k0l'] / line['bi1.bsw1l1.1'].length)
    line.vars['k0bi1bsw1l12'] = (-line.vars['bsw_k0l'] / line['bi1.bsw1l1.2'].length)
    line.vars['k0bi1bsw1l13'] = (-line.vars['bsw_k0l'] / line['bi1.bsw1l1.3'].length)
    line.vars['k0bi1bsw1l14'] = (line.vars['bsw_k0l'] / line['bi1.bsw1l1.4'].length)


    # Build knob to model eddy currents (k2)
    line.vars['bsw_k2l'] = 0
    line.element_refs['bi1.bsw1l1.1'].knl[2] = line.vars['bsw_k2l']
    line.element_refs['bi1.bsw1l1.2'].knl[2] = -line.vars['bsw_k2l']
    line.element_refs['bi1.bsw1l1.3'].knl[2] = -line.vars['bsw_k2l']
    line.element_refs['bi1.bsw1l1.4'].knl[2] = line.vars['bsw_k2l']

    # Test to_dict roundtrip
    line = xt.Line.from_dict(line.to_dict())
    line.build_tracker(_context=test_context)

    bsw_k2l_ref = -9.7429e-2
    bsw_k0l_ref = 6.6e-2

    line.vars['bsw_k2l'] = bsw_k2l_ref
    line.vars['bsw_k0l'] = bsw_k0l_ref
    xo.assert_allclose(line['bi1.bsw1l1.1']._xobject.knl[2], bsw_k2l_ref, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.2']._xobject.knl[2], -bsw_k2l_ref, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.3']._xobject.knl[2], -bsw_k2l_ref, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.4']._xobject.knl[2], bsw_k2l_ref, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.1'].k0, bsw_k0l_ref / line['bi1.bsw1l1.1'].length, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.2'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.2'].length, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.3'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.3'].length, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.4'].k0, bsw_k0l_ref / line['bi1.bsw1l1.4'].length, rtol=0, atol=1e-10)

    tw = line.twiss()
    xo.assert_allclose(tw['x', 'bi1.tstr1l1'], -0.045739596, rtol=0, atol=1e-5)
    xo.assert_allclose(tw['y', 'bi1.tstr1l1'], 0.0000000, rtol=0, atol=1e-5)
    xo.assert_allclose(tw['betx', 'bi1.tstr1l1'], 5.201734, rtol=0, atol=1e-4)
    xo.assert_allclose(tw['bety', 'bi1.tstr1l1'], 6.900674, rtol=0, atol=1e-4)
    xo.assert_allclose(tw.qy, 4.474426935973243, rtol=0, atol=1e-6) # verify that it does not change from one version to the other
    xo.assert_allclose(tw.qx, 4.396759378318215, rtol=0, atol=1e-6)
    xo.assert_allclose(tw.dqy, -8.624431, rtol=0, atol=1e-3)
    xo.assert_allclose(tw.dqx, -3.560640, rtol=0, atol=1e-3)

    line.vars['bsw_k2l'] = bsw_k2l_ref / 3
    xo.assert_allclose(line['bi1.bsw1l1.1']._xobject.knl[2], bsw_k2l_ref / 3, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.2']._xobject.knl[2], -bsw_k2l_ref / 3, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.3']._xobject.knl[2], -bsw_k2l_ref / 3, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.4']._xobject.knl[2], bsw_k2l_ref / 3, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.1'].k0, bsw_k0l_ref / line['bi1.bsw1l1.1'].length, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.2'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.2'].length, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.3'].k0, -bsw_k0l_ref / line['bi1.bsw1l1.3'].length, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.4'].k0, bsw_k0l_ref / line['bi1.bsw1l1.4'].length, rtol=0, atol=1e-10)

    tw = line.twiss()
    xo.assert_allclose(tw['x', 'bi1.tstr1l1'], -0.045887, rtol=0, atol=1e-5)
    xo.assert_allclose(tw['y', 'bi1.tstr1l1'], 0.0000000, rtol=0, atol=1e-5)
    xo.assert_allclose(tw['betx', 'bi1.tstr1l1'], 5.264522, rtol=0, atol=1e-4)
    xo.assert_allclose(tw['bety', 'bi1.tstr1l1'], 6.317935, rtol=0, atol=1e-4)
    xo.assert_allclose(tw.qy, 4.4717791778, rtol=0, atol=1e-6)
    xo.assert_allclose(tw.qx, 4.3989420079, rtol=0, atol=1e-6)
    xo.assert_allclose(tw.dqy, -8.2045492725, rtol=0, atol=1e-3)
    xo.assert_allclose(tw.dqx, -3.5636322837, rtol=0, atol=1e-3)

    # Switch off bsws
    line.vars['bsw_k0l'] = 0
    line.vars['bsw_k2l'] = 0
    xo.assert_allclose(line['bi1.bsw1l1.1']._xobject.knl[2], 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.2']._xobject.knl[2], 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.3']._xobject.knl[2], 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.4']._xobject.knl[2], 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.1'].k0, 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.2'].k0, 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.3'].k0, 0, rtol=0, atol=1e-10)
    xo.assert_allclose(line['bi1.bsw1l1.4'].k0, 0, rtol=0, atol=1e-10)

    tw = line.twiss()
    xo.assert_allclose(tw['x', 'bi1.tstr1l1'], 0, rtol=0, atol=1e-5)
    xo.assert_allclose(tw['y', 'bi1.tstr1l1'], 0, rtol=0, atol=1e-5)
    xo.assert_allclose(tw['betx', 'bi1.tstr1l1'], 5.2996347, rtol=0, atol=1e-4)
    xo.assert_allclose(tw['bety', 'bi1.tstr1l1'], 3.838857, rtol=0, atol=1e-4)
    xo.assert_allclose(tw.qy, 4.45, rtol=0, atol=1e-6)
    xo.assert_allclose(tw.qx, 4.4, rtol=0, atol=1e-6)
    xo.assert_allclose(tw.dqy, -7.14978134184, rtol=0, atol=1e-3)
    xo.assert_allclose(tw.dqx, -3.56557575115, rtol=0, atol=1e-3)


    # Setup time-dependent functions

    df = pd.read_csv(test_data_folder /
                     'psb_chicane/chicane_collapse.csv',
                     delimiter=',', skipinitialspace=True)

    line.functions['fun_bsw_k0l'] = xd.FunctionPieceWiseLinear(
        x=df['time'].values, y=df['bsw_k0l'].values)
    line.functions['fun_bsw_k2l'] = xd.FunctionPieceWiseLinear(
        x=df['time'].values, y=df['bsw_k2l'].values)

    # Control knob with function
    line.vars['on_chicane_k0'] = 1
    line.vars['on_chicane_k2'] = 1
    line.vars['bsw_k0l'] = (line.functions.fun_bsw_k0l(line.vars['t_turn_s'])
                            * line.vars['on_chicane_k0'])
    line.vars['bsw_k2l'] = (line.functions.fun_bsw_k2l(line.vars['t_turn_s'])
                            * line.vars['on_chicane_k2'])


    # Test to_dict roundtrip
    line = xt.Line.from_dict(line.to_dict())
    line.build_tracker(_context=test_context)

    # Correct tunes and beta beat
    line.discard_tracker()
    line.insert_element(element=xt.Marker(), name='mker_match', at_s=79.874)
    line.build_tracker(_context=test_context)

    line.vars['on_chicane_k0'] = 0
    line.vars['on_chicane_k2'] = 0
    tw0 = line.twiss()
    line.vars['on_chicane_k0'] = 1
    line.vars['on_chicane_k2'] = 1


    t_correct = np.linspace(0, 5.5e-3, 30)

    kbrqf_corr_list = []
    kbrqd_corr_list = []
    kbrqd3corr_list = []
    kbrqd14corr_list = []
    for ii, tt in enumerate(t_correct):
        print(f'Correct tune at t = {tt * 1e3:.2f} ms   \n')
        line.vars['t_turn_s'] = tt

        line.match(
            #verbose=True,
            vary=[
                xt.Vary('kbrqfcorr', step=1e-4),
                xt.Vary('kbrqdcorr', step=1e-4),
                xt.Vary('kbrqd3corr', step=1e-4),
                xt.Vary('kbrqd14corr', step=1e-4),
            ],
            targets = [
                xt.Target('qx', value=tw0.qx, tol=1e-5, scale=1),
                xt.Target('qy', value=tw0.qy, tol=1e-5, scale=1),
                xt.Target('bety', at='mker_match',
                        value=tw0['bety', 'mker_match'], tol=1e-4, scale=100),
                xt.Target('alfy', at='mker_match',
                        value=tw0['alfy', 'mker_match'], tol=1e-4, scale=100)
            ]
        )

        kbrqf_corr_list.append(line.vars['kbrqfcorr']._value)
        kbrqd_corr_list.append(line.vars['kbrqdcorr']._value)
        kbrqd3corr_list.append(line.vars['kbrqd3corr']._value)
        kbrqd14corr_list.append(line.vars['kbrqd14corr']._value)

    line.functions['fun_kqf_corr'] = xd.FunctionPieceWiseLinear(
        x=t_correct, y=kbrqf_corr_list)
    line.functions['fun_kqd_corr'] = xd.FunctionPieceWiseLinear(
        x=t_correct, y=kbrqd_corr_list)
    line.functions['fun_qd3_corr'] = xd.FunctionPieceWiseLinear(
        x=t_correct, y=kbrqd3corr_list)
    line.functions['fun_qd14_corr'] = xd.FunctionPieceWiseLinear(
        x=t_correct, y=kbrqd14corr_list)

    line.vars['on_chicane_tune_corr'] = 1
    line.vars['kbrqfcorr'] = (line.vars['on_chicane_tune_corr']
                                * line.functions.fun_kqf_corr(line.vars['t_turn_s']))
    line.vars['kbrqdcorr'] = (line.vars['on_chicane_tune_corr']
                                * line.functions.fun_kqd_corr(line.vars['t_turn_s']))

    line.vars['on_chicane_beta_corr'] = 1
    line.vars['kbrqd3corr'] = (line.vars['on_chicane_beta_corr']
                            * line.functions.fun_qd3_corr(line.vars['t_turn_s']))
    line.vars['kbrqd14corr'] = (line.vars['on_chicane_beta_corr']
                            * line.functions.fun_qd14_corr(line.vars['t_turn_s']))

    line.vars['on_chicane_k0'] = 1
    line.vars['on_chicane_k2'] = 1
    line.vars['on_chicane_beta_corr'] = 0
    line.vars['on_chicane_tune_corr'] = 0

    line_thick = line.copy()
    line_thick.build_tracker(_context=test_context)

    line.discard_tracker()

    slicing_strategies = [
        Strategy(slicing=Teapot(1)),  # Default
        Strategy(slicing=Teapot(2), element_type=xt.Bend),
        Strategy(slicing=Teapot(2), element_type=xt.RBend),
        Strategy(slicing=Teapot(8), element_type=xt.Quadrupole),
    ]

    print("Slicing thick elements...")
    line.slice_thick_elements(slicing_strategies)

    line.build_tracker(_context=test_context)

    tw_thin = line.twiss()
    tw_thick = line_thick.twiss()

    print('\n')
    print(f'Qx: thick {tw_thin.qx:.4f} thin {tw_thick.qx:.4f}, diff {tw_thin.qx-tw_thick.qx:.4e}')
    print(f'Qy: thick {tw_thin.qy:.4f} thin {tw_thick.qy:.4f}, diff {tw_thin.qy-tw_thick.qy:.4e}')
    print(f"Q'x: thick {tw_thin.dqx:.4f} thin {tw_thick.dqx:.4f}, diff {tw_thin.dqx-tw_thick.dqx:.4f}")
    print(f"Q'y: thick {tw_thin.dqy:.4f} thin {tw_thick.dqy:.4f}, diff {tw_thin.dqy-tw_thick.dqy:.4f}")

    bety_interp = np.interp(tw_thick.s, tw_thin.s, tw_thin.bety)
    print(f"Max beta beat: {np.max(np.abs(tw_thick.bety/bety_interp - 1)):.4e}")

    #####################
    # Check against ptc #
    #####################

    with open(test_data_folder / 'psb_chicane/ptc_ref.json', 'r') as fid: # generated by xtrack/examples/psb/001_compare_against_ptc.py
        ptc_ref = json.load(fid)

    for kk, vv in ptc_ref.items():
        ptc_ref[kk] = np.array(vv)

    t_test = ptc_ref['t_test']
    qx_ptc = ptc_ref['qx_ptc']
    qy_ptc = ptc_ref['qy_ptc']
    dqx_ptc = ptc_ref['dqx_ptc']
    dqy_ptc = ptc_ref['dqy_ptc']
    bety_at_scraper_ptc = ptc_ref['bety_at_scraper_ptc']

    # Check against ptc (correction off)
    line.vars['on_chicane_beta_corr'] = 0
    line.vars['on_chicane_tune_corr'] = 0
    qx_thick = []
    qy_thick = []
    dqx_thick = []
    dqy_thick = []
    bety_at_scraper_thick = []
    qx_thin = []
    qy_thin = []
    dqx_thin = []
    dqy_thin = []
    bety_at_scraper_thin = []
    for ii, tt in enumerate(t_test):
        print(f'Check against ptc, twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
        line_thick.vars['t_turn_s'] = tt
        line.vars['t_turn_s'] = tt

        tw_thick = line_thick.twiss()
        bety_at_scraper_thick.append(tw_thick['bety', 'br.stscrap22'])
        qx_thick.append(tw_thick.qx)
        qy_thick.append(tw_thick.qy)
        dqx_thick.append(tw_thick.dqx)
        dqy_thick.append(tw_thick.dqy)

        tw_thin = line.twiss()
        bety_at_scraper_thin.append(tw_thin['bety', 'br.stscrap22'])
        qx_thin.append(tw_thin.qx)
        qy_thin.append(tw_thin.qy)
        dqx_thin.append(tw_thin.dqx)
        dqy_thin.append(tw_thin.dqy)


    xo.assert_allclose(qx_thick, qx_ptc, atol=2e-4, rtol=0)
    xo.assert_allclose(qy_thick, qy_ptc, atol=2e-4, rtol=0)
    xo.assert_allclose(qx_thin, qx_ptc, atol=1e-3, rtol=0)
    xo.assert_allclose(qy_thin, qy_ptc, atol=1e-3, rtol=0)
    xo.assert_allclose(dqx_thick, dqx_ptc, atol=0.5, rtol=0)
    xo.assert_allclose(dqy_thick, dqy_ptc, atol=0.5, rtol=0)
    xo.assert_allclose(dqx_thin, dqx_ptc, atol=0.5, rtol=0)
    xo.assert_allclose(dqy_thin, dqy_ptc, atol=0.5, rtol=0)
    xo.assert_allclose(bety_at_scraper_thick, bety_at_scraper_ptc, atol=0, rtol=1e-2)
    xo.assert_allclose(bety_at_scraper_thin, bety_at_scraper_ptc, atol=0, rtol=2e-2)

    # Check correction
    line.vars['on_chicane_beta_corr'] = 1
    line.vars['on_chicane_tune_corr'] = 1
    line_thick.vars['on_chicane_beta_corr'] = 1
    line_thick.vars['on_chicane_tune_corr'] = 1
    qx_thick_corr = []
    qy_thick_corr = []
    bety_at_scraper_thick_corr = []
    qx_thin_corr = []
    qy_thin_corr = []
    bety_at_scraper_thin_corr = []
    for ii, tt in enumerate(t_test):
        print(f'Check correction, twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
        line_thick.vars['t_turn_s'] = tt
        line.vars['t_turn_s'] = tt

        tw_thick = line_thick.twiss()
        bety_at_scraper_thick_corr.append(tw_thick['bety', 'br.stscrap22'])
        qx_thick_corr.append(tw_thick.qx)
        qy_thick_corr.append(tw_thick.qy)

        tw_thin = line.twiss()
        bety_at_scraper_thin_corr.append(tw_thin['bety', 'br.stscrap22'])
        qx_thin_corr.append(tw_thin.qx)
        qy_thin_corr.append(tw_thin.qy)

    qx_thick_corr = np.array(qx_thick_corr)
    qy_thick_corr = np.array(qy_thick_corr)
    bety_at_scraper_thick_corr = np.array(bety_at_scraper_thick_corr)
    qx_thin_corr = np.array(qx_thin_corr)
    qy_thin_corr = np.array(qy_thin_corr)
    bety_at_scraper_thin_corr = np.array(bety_at_scraper_thin_corr)

    xo.assert_allclose(qx_thick_corr, qx_ptc[-1], atol=3e-3, rtol=0)
    xo.assert_allclose(qy_thick_corr, qy_ptc[-1], atol=3e-3, rtol=0)
    xo.assert_allclose(qx_thin_corr, qx_ptc[-1], atol=3e-3, rtol=0)
    xo.assert_allclose(qy_thin_corr, qy_ptc[-1], atol=3e-3, rtol=0)
    xo.assert_allclose(bety_at_scraper_thick_corr, bety_at_scraper_ptc[-1], atol=0, rtol=3e-2)
    xo.assert_allclose(bety_at_scraper_thin_corr, bety_at_scraper_ptc[-1], atol=0, rtol=3e-2)

    # Tracking with time-dependent variable

    line.vars['on_chicane_k0'] = 1
    line.vars['on_chicane_k2'] = 1
    line.vars['on_chicane_beta_corr'] = 1
    line.vars['on_chicane_tune_corr'] = 1

    # Install monitor at foil
    monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=6000, num_particles=1)
    line.discard_tracker()
    line.insert_element(index='bi1.tstr1l1', element=monitor, name='monitor_at_foil')
    line.build_tracker(_context=test_context)

    p = line.build_particles(x=0, px=0, y=0, py=0, delta=0, zeta=0)

    line.enable_time_dependent_vars = True
    line.dt_update_time_dependent_vars = 3e-6

    print('Tracking...')
    line.track(p, num_turns=6000, time=True)
    print(f'Done in {line.time_last_track:.4} s')

    xo.assert_allclose(monitor.x[0, 0], -0.045936, rtol=0, atol=1e-5)
    xo.assert_allclose(monitor.x[0, 300], -0.04522354, rtol=0, atol=1e-5)
    xo.assert_allclose(monitor.x[0, 2500], -0.02256763, rtol=0, atol=1e-5)
    xo.assert_allclose(monitor.x[0, 4500], -0.00143883, rtol=0, atol=1e-5)
    xo.assert_allclose(monitor.x[0, 5500], 0.0, rtol=0, atol=1e-5)

    # Test multiturn injection
    if isinstance(test_context, xo.ContextCpu):
        line.t_turn_s = 0 # Reset time!

        line.vars['on_chicane_k0'] = 1
        line.vars['on_chicane_k2'] = 1
        line.vars['on_chicane_beta_corr'] = 1
        line.vars['on_chicane_tune_corr'] = 1

        df = pd.read_table(test_data_folder / 'psb_chicane/inj_distrib.dat',
            skiprows=3,
            names="x x' y y' z z' Phase Time Energy Loss".split())

        kin_energy_ev = df.Energy.values * 1e6
        tot_energy_ev = kin_energy_ev + xt.PROTON_MASS_EV
        p0c = line.particle_ref.p0c[0]
        tot_energy0_ev = line.particle_ref.energy0[0]
        ptau = (tot_energy_ev - tot_energy0_ev) / p0c

        part_for_injection = xt.Particles(q0=1, mass0=xt.PROTON_MASS_EV, p0c=line.particle_ref.p0c[0],
                                        ptau=ptau)

        part_for_injection.x = df.x.values * 1e-3
        part_for_injection.y = df.y.values * 1e-3
        part_for_injection.zeta = df.z.values * 1e-3
        part_for_injection.px = df["x'"].values  * 1e-3 * (1 + part_for_injection.delta)
        part_for_injection.py = df["y'"].values  * 1e-3 * (1 + part_for_injection.delta)
        part_for_injection.weight = 10

        p_injection = xt.ParticlesInjectionSample(particles_to_inject=part_for_injection,
                                                line=line,
                                                element_name='injection',
                                                num_particles_to_inject=7)

        line.discard_tracker()
        line.insert_element(index='bi1.tstr1l1', element=p_injection, name='injection')

        # Add monitor at end line
        monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=10, num_particles=100)
        line.insert_element(index='p16ring1$end', element=monitor, name='monitor_at_end')
        line.build_tracker()

        p = line.build_particles(_capacity=100, x=0)
        p.state[0] = -500 # kill the particle added by default

        intensity = []
        line.enable_time_dependent_vars = True
        for iturn in range(8):
            intensity.append(p.weight[p.state>0].sum())
            line.track(p, num_turns=1)

        assert np.all(np.sum(monitor.state > 0, axis=0)
                    == np.array([ 0,  7, 14, 21, 28, 35, 42, 49, 56,  0]))

        assert np.all(monitor.at_element[monitor.state > 0] ==
                    line.element_names.index('monitor_at_end'))

        xo.assert_allclose(monitor.s[monitor.state > 0],
                        line.get_s_position('monitor_at_end'), atol=1e-12)



import numpy as np
import xtrack as xt

# Get a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

# Switch RF on
line.vars['vrf400'] = 16
line.vars['lagrf400.b1'] = 0.5

# Enable crossing angle orbit bumps
line.vars['on_x1'] = 10
line.vars['on_x2'] = 20
line.vars['on_x5'] = 10
line.vars['on_x8'] = 30

# Generate line made on maps (splitting at defined markers)
ele_cut = ['ip1', 'ip2', 'ip5', 'ip8'] # markers where to split the line
line_maps = line.get_line_with_second_order_maps(split_at=ele_cut)
line_maps.build_tracker()

line_maps.get_table().show()
# prints:
#
# name              s element_type         isthick
# ip7               0 Marker                 False
# map_0             0 SecondOrderTaylorMap    True
# ip8         3321.22 Marker                 False
# map_1       3321.22 SecondOrderTaylorMap    True
# ip1         6664.72 Marker                 False
# map_2       6664.72 SecondOrderTaylorMap    True
# ip2         9997.16 Marker                 False
# map_3       9997.16 SecondOrderTaylorMap    True
# ip5           19994 Marker                 False
# map_4         19994 SecondOrderTaylorMap    True
# lhcb1ip7_p_ 26658.9 Marker                 False
# _end_point  26658.9                        False

# Compare twiss of the two lines
tw = line.twiss()
tw_map = line_maps.twiss()

tw.qx       # is 62.3099999
tw_map.qx   # is  0.3099999

tw.dqx      # is 1.9135
tw_map.dqx  # is 1.9137

tw.rows[['ip1', 'ip2', 'ip5', 'ip8']].cols['x px y py betx bety'].show()
# prints
#
# name           x          px           y          py    betx    bety
# ip8  9.78388e-09 3.00018e-05 6.25324e-09 5.47624e-09 1.50001 1.49999
# ip1  1.92278e-10 1.00117e-05 2.64869e-09  1.2723e-08    0.15    0.15
# ip2  1.47983e-08 2.11146e-09 -2.1955e-08 2.00015e-05      10 9.99997
# ip5 -3.03593e-09 5.75413e-09 5.13184e-10  1.0012e-05    0.15    0.15

tw_map.rows[['ip1', 'ip2', 'ip5', 'ip8']].cols['x px y py betx bety']
# prints
#
# name             x          px           y          py    betx    bety
# ip8    9.78388e-09 3.00018e-05 6.25324e-09 5.47624e-09 1.50001 1.49999
# ip1    1.92278e-10 1.00117e-05 2.64869e-09  1.2723e-08    0.15    0.15
# ip2    1.47983e-08 2.11146e-09 -2.1955e-08 2.00015e-05      10 9.99997
# ip5   -3.03593e-09 5.75413e-09 5.13184e-10  1.0012e-05    0.15    0.15


import numpy as np
from scipy.constants import c as clight

from .general import _print

import xtrack as xt
import xobjects as xo

def compensate_radiation_energy_loss(line, delta0=0, rtol_eneloss=1e-12,
                                     max_iter=100, verbose=True, **kwargs):

    assert isinstance(line._context, xo.ContextCpu), "Only CPU context is supported"
    assert line.particle_ref is not None, "Particle reference is not set"
    assert np.abs(line.particle_ref.q0) == 1, "Only |q0| = 1 is supported (for now)"

    if 'record_iterations' in kwargs:
        record_iterations = kwargs['record_iterations']
        kwargs.pop('record_iterations')
        line._tapering_iterations = []
    else:
        record_iterations = False

    if verbose: _print("Compensating energy loss.")

    line.config.XTRACK_MULTIPOLE_NO_SYNRAD = True
    with xt.freeze_longitudinal(line):
        particle_on_co = line.find_closed_orbit()
    line.config.XTRACK_MULTIPOLE_NO_SYNRAD = False

    beta0 = float(particle_on_co._xobject.beta0[0])

    # Check whether compensation is needed
    p_test = particle_on_co.copy()
    p_test.delta = delta0
    line.track(p_test, turn_by_turn_monitor='ONE_TURN_EBE')
    mon = line.record_last_track
    eloss = -(mon.ptau[0, -1] - mon.ptau[0, 0]) * p_test.p0c[0]
    if p_test.state[0] > 0 and abs(eloss) < p_test.energy0[0] * rtol_eneloss:
        if verbose: _print("  - No compensation needed")
        return

    # save voltages
    v_setter = line.attr._cache['voltage'].multisetter
    f_setter = line.attr._cache['frequency'].multisetter
    lag_setter = line.attr._cache['lag'].multisetter
    lag_taper_setter = line.attr._cache['lag_taper'].multisetter

    v0 = v_setter.get_values()
    f0 = f_setter.get_values()
    lag_zero = lag_setter.get_values()

    eneloss_partitioning = v0 / v0.sum()

    # Put all cavities on crest and at zero frequency
    lag_taper_setter.set_values(90. - lag_zero)
    v_setter.set_values(np.zeros_like(v_setter.get_values()))
    f_setter.set_values(np.zeros_like(f_setter.get_values()))

    if verbose: _print("Share energy loss among cavities (repeat until energy loss is zero)")
    with xt.line._preserve_config(line):
        line.config.XTRACK_MULTIPOLE_TAPER = True
        line.config.XTRACK_DIPOLEEDGE_TAPER = True

        i_iter = 0
        while True:
            p_test = particle_on_co.copy()
            p_test.delta = delta0
            line.track(p_test, turn_by_turn_monitor='ONE_TURN_EBE')
            mon = line.record_last_track

            if record_iterations:
                line._tapering_iterations.append(mon)

            eloss = -(mon.ptau[0, -1] - mon.ptau[0, 0]) * p_test.p0c[0]
            if verbose: _print(f"Energy loss: {eloss:.3f} eV             ")#, end='\r', flush=True)

            if eloss < p_test.energy0[0] * rtol_eneloss:
                break

            v_setter.set_values(v_setter.get_values()
                                + eloss * eneloss_partitioning)

            i_iter += 1
            if i_iter > max_iter:
                raise RuntimeError("Maximum number of iterations reached")
    if verbose: _print()
    delta_taper_full = 0.5*(mon.delta[0, :-1] + mon.delta[0, 1:]) # last point is added by the monitor

    if verbose: _print("  - Set delta_taper")
    delta_taper_mask = line.attr._cache['delta_taper'].mask
    delta_taper_setter = line.attr._cache['delta_taper'].multisetter
    delta_taper_setter.set_values(delta_taper_full[delta_taper_mask])

    if verbose: _print("  - Restore cavity voltage and frequency. Set cavity lag")
    v_synchronous = v_setter.get_values()

    mask_cav = line.attr._cache['voltage'].mask
    zeta_at_cav = np.atleast_1d(np.squeeze(mon.zeta[0, :-1]))[mask_cav]
    mask_active_cav = np.abs(v0) > 0
    v_ratio = v0 * 0
    v_ratio[mask_active_cav] = v_synchronous[mask_active_cav] / v0[mask_active_cav]
    assert np.all(np.abs(v_ratio[mask_active_cav]) < 1)
    inst_phase = np.arcsin(v_ratio)

    total_lag = 360.*(inst_phase / (2 * np.pi) - f0 * zeta_at_cav / beta0 / clight)
    total_lag = 180. - total_lag # we are above transition
    lag_taper = total_lag - lag_zero
    lag_taper[~mask_active_cav] = 0

    v_setter.set_values(v0)
    f_setter.set_values(f0)
    lag_taper_setter.set_values(lag_taper)




import xtrack as xt

env = xt.Environment()

# Create some quadrupole elements
env['kquad'] = 0.1
env.new('q1', xt.Quadrupole, length=1.0, k1='kquad')
env.new('q2', xt.Quadrupole, length=1.0, k1='-kquad')
env.new('q3', xt.Quadrupole, length=1.0, k1='kquad')
env.new('q4', xt.Quadrupole, length=1.0, k1='-kquad')
env.new('s4', xt.Sextupole, length=0.1)

myline = env.new_line(name='myline', components=[
    # Place element center at s = 3.0
    env.place('q1', at=3.0),
    # Place element start at s = 5.0
    env.place('q2', anchor='start', at=5.0),
    # Place element start at the end of q2
    env.place('q3', anchor='start', at='q2@end'),
    # Place element center at 5 m from the end of q3
    env.place('q4', anchor='center', at=5.0, from_='q3@start'),
    # Placed right after previous element
    env.place('s4')
    ])

tt = myline.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0          1.25           2.5
# q1                   2.5             3           3.5
# drift_2              3.5          4.25             5
# q2                     5           5.5             6
# q3                     6           6.5             7
# drift_3                7          8.75          10.5
# q4                  10.5            11          11.5
# s4                  11.5         11.55          11.6
# _end_point          11.6          11.6          11.6


# The elements can also be created directly in the line definition:
myline = env.new_line(name='myline', components=[
    env.new('q10', xt.Quadrupole, length=1.0, k1='kquad',
            # Place element center at s = 3.0
            at=3.0),
    env.new('q20', xt.Quadrupole, length=1.0, k1='-kquad',
            # Place element start at s = 5.0
            anchor='start', at=5.0),
    env.new('q30', xt.Quadrupole, length=1.0, k1='kquad',
            # Place element start at the end of q2
            anchor='start', at='q20@end'),
    env.new('q40', xt.Quadrupole, length=1.0, k1='-kquad',
            # Place element center at the end of q3
            anchor='center', at=5.0, from_='q30@start'),
    env.new('s40', xt.Sextupole, length=0.1) # Placed right after previous
    ])

tt = myline.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_4                0          1.25           2.5
# q10                  2.5             3           3.5
# drift_5              3.5          4.25             5
# q20                    5           5.5             6
# q30                    6           6.5             7
# drift_6                7          8.75          10.5
# q40                 10.5            11          11.5
# s40                 11.5         11.55          11.6
# _end_point          11.6          11.6          11.6


import xtrack as xt

import xtrack._temp.lhc_match as lm

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()
collider.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

default_tol = {None: 1e-8, 'betx': 1e-6, 'bety': 1e-6} # to have no rematching w.r.t. madx

tw_sq_ip1_b1 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
        line_name='lhcb1', start='s.ds.r8.b1', end='e.ds.l2.b1',
        beta_star_x=collider.varval['betxip1b1'],
        beta_star_y=collider.varval['betyip1b1'])

mux_compensate = (tw_sq_ip1_b1['mux', 's.ds.l2.b1'] - tw_sq_ip1_b1['mux', 'ip1']
                   - collider.varval['muxip1b1_r'] - collider.varval['mux12b1'])
mux_ir2_target = collider.varval['muxip2b1'] - mux_compensate
muy_compensate = (tw_sq_ip1_b1['muy', 's.ds.l2.b1'] - tw_sq_ip1_b1['muy', 'ip1']
                  - collider.varval['muyip1b1_r'] - collider.varval['muy12b1'])
muy_ir2_target = collider.varval['muyip2b1'] - muy_compensate

arc_periodic_solution = lm.get_arc_periodic_solution(collider)

opt = collider.lhcb1.match(
    solve=False,
    default_tol=default_tol,
    start='s.ds.l2.b1', end='e.ds.r2.b1',
    # Left boundary
    init=tw_sq_ip1_b1,
    targets=[
        # IP optics
        xt.TargetSet(at='ip2',
            betx=collider.varval['betxip2b1'], bety=collider.varval['betyip2b1'],
            alfx=0, alfy=0, dx=0, dpx=0),
        # Right boundary
        xt.TargetList(('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                value=arc_periodic_solution['lhcb1']['23'], at='e.ds.r2.b1'),
        xt.TargetRelPhaseAdvance('mux', mux_ir2_target),
        xt.TargetRelPhaseAdvance('muy', muy_ir2_target),
    ],
    vary=xt.VaryList(
        ['kq4.l2b1', 'kq5.l2b1',  'kq6.l2b1', 'kq7.l2b1', 'kq8.l2b1',
         'kq4.r2b1', 'kq5.r2b1',  'kq6.r2b1', 'kq7.r2b1', 'kq8.r2b1',
         'kq9.l2b1', 'kq10.l2b1', 'kqtl11.l2b1', 'kqt12.l2b1', 'kqt13.l2b1',
         'kq9.r2b1', 'kq10.r2b1', 'kqtl11.r2b1', 'kqt12.r2b1', 'kqt13.r2b1']
    )
)

opt.solve()


import bpmeth
import xtrack as xt
import sympy

import numpy as np

b1 = "0.2"
b2 = "0.1"
h = "0.1"
length = 0.5

s = sympy.symbols('s')
b1_f = sympy.lambdify(s, b1, modules='numpy')

# Vector potential (bs, b1, a1, b2, a2, b3, a3, ,... as function of s)
A_magnet = bpmeth.GeneralVectorPotential(hs=h, b=(b1,b2))

class MyHamiltonian(bpmeth.Hamiltonian):
    def get_H(self, coords):
        x, y, s = coords.x, coords.y, coords.s
        px, py, ptau = coords.px, coords.py, coords.ptau
        beta0 = coords.beta0
        h = self.curv
        one_plus_delta_allsq = 1+2*ptau/beta0+ptau**2
        sqrt = coords._m.sqrt
        term1 = ptau / beta0
        term2 = 0.5 * (px**2 + py**2) / sqrt(one_plus_delta_allsq)
        term3 = (float(b1) - h) * x + float(b1) * h * x**2 / 2
        term4 = float(b2) * (x**2 - y**2) / 2

        term_correction = float(b2) * h * (2*x**3 - 3*x*y**2) / 6

        H = term1 + term2 + term3 + term4 + term_correction
        return H

# Make Hamiltonian for a defined length
H_magnet = MyHamiltonian(length=length, curv=float(h), vectp=A_magnet)

# This guy is able to track an Xsuite particle!
p0 = xt.Particles(x=np.linspace(-1e-3, 1e-3, 10),
                  y=np.linspace(-2e-3, 2e-3, 10),
                   energy0=10e9, mass0=xt.ELECTRON_MASS_EV)

p_silke = p0.copy()
sol = H_magnet.track(p_silke, return_sol=True,
                     ivp_opt= {"rtol":1e-10, "atol":1e-12})

p_xsuite = p0.copy()
bb = xt.Bend(h=float(h), length=length, k0=float(b1), k1=float(b2))
bb.edge_entry_active = False
bb.edge_exit_active = False

bb.num_multipole_kicks = 100
bb.integrator = 'yoshida4'

bb.model =  'bend-kick-bend'
bb.model =  'mat-kick-mat'

bb.track(p_xsuite)

# Print the max difference
print(f'Max difference on x: {np.max(np.abs(p_silke.x - p_xsuite.x))}')

# I want to see the trajectory along the object
import matplotlib.pyplot as plt
plt.close('all')
plt.figure()
for ss in sol:
    plt.plot(ss.t, ss.y[0])
plt.xlabel('s [m]')
plt.ylabel('x [m]')

plt.show()



import json
import io
from pathlib import Path
import gzip
from xobjects import JEncoder


def dump(data, file, indent=1):
    if isinstance(file, io.IOBase):
        fh, close = file, False
    elif (isinstance(file, str) and file.endswith(".gz")) or (
        isinstance(file, Path) and file.suffix == ".gz"
    ):
        fh, close = gzip.open(file, "wt"), True
    else:
        fh, close = open(file, "w"), True

    json.dump(data, fh, indent=indent, cls=JEncoder)

    if close:
        fh.close()


def load(file):
    if isinstance(file, io.IOBase):
        fh, close = file, False
    elif (isinstance(file, str) and file.endswith(".gz")) or (
        isinstance(file, Path) and file.suffix == ".gz"
    ):
        fh, close = gzip.open(file, "rt"), True
    else:
        fh, close = open(file, "r"), True

    data = json.load(fh)

    if close:
        fh.close()

    return data



import xtrack as xt

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider['lhcb1'].twiss_default['method'] = '4d'
collider['lhcb2'].twiss_default['method'] = '4d'

line = collider.lhcb1
line.cycle('ip3', inplace=True)

tw = line.twiss()

init = xt.TwissInit(betx=0.15, bety=0.15,
                           ax_chrom=42.7928, bx_chrom=-18.4181,
                           ay_chrom=-18.0191, by_chrom= 11.54862)
tw_open = line.twiss(start='ip5', end='ip7', init=init,
                        compute_chromatic_properties=True,
                        )

ksfs = ['ksf1.a81b1', 'ksf1.a12b1', 'ksf1.a45b1', 'ksf1.a56b1']
ksds = ['ksd2.a81b1', 'ksd2.a12b1', 'ksd2.a45b1', 'ksd2.a56b1']

for kk in ksfs:
    line.vv[kk] = 0.06
for kk in ksds:
    line.vv[kk] = -0.099

twl = line.twiss()

opt = line.match(solve=False,
        vary=[xt.VaryList(
            ('ksf1.a81b1 ksd2.a81b1 ksf1.a12b1 ksd2.a12b1 ksf1.a45b1 '
             'ksd2.a45b1 ksf1.a56b1 ksd2.a56b1').split(),
            limits=(-0.38, 0.38), step=1e-6)],
        targets=[
            xt.Target('ax_chrom', 0, at='ip3'),
            xt.Target('bx_chrom', 0, at='ip3'),
            xt.Target('ay_chrom', 0, at='ip3'),
            xt.Target('by_chrom', 0, at='ip3'),
            xt.Target(lambda tt: tt['wx_chrom', 'ip3'] - tt['wx_chrom', 'ip7'], 0),
            xt.Target(lambda tt: tt['wy_chrom', 'ip3'] - tt['wy_chrom', 'ip7'], 0),
            xt.Target('bx_chrom', 0, at='ip1'),
            xt.Target('by_chrom', 0, at='ip1'),
            xt.Target('bx_chrom', 0, at='ip5'),
            xt.Target('by_chrom', 0, at='ip5'),
        ])


import matplotlib.pyplot as plt
plt.close('all')

plt.plot(tw.s, tw.dmux)

for nn in tw.rows['mbxf.*'].name:
    plt.axvline(tw['s', nn], color='k', linestyle='--')

plt.show()


import xtrack as xt
from cpymad.madx import Madx

# Load a very simple sequence from MAD-X
mad = Madx()
mad.input("""
    seq: sequence, l=4;
    b1: sbend, at=0.5, angle=0.2, l=1;
    q1: quadrupole, at=2.5, k1=0.1, l=1;
    endsequence;

    beam;
    use,sequence=seq;
""")

line = xt.Line.from_madx_sequence(mad.sequence.seq)
line.build_tracker()

print('The line as imported from MAD-X:')
line.get_table().show()

# Add multipolar components to elements
line['b1'].knl[2] = 0.001 # Normal sxtupole component
line['q1'].ksl[3] = 0.002 # Skew octupole component

tt = line.get_table(attr=True)
tt.cols['s', 'element_type', 'isthick', 'k2l', 'k3sl'].show()
# returns:
#
# name       s element_type isthick   k2l  k3sl
# seq$start  0 Marker         False     0     0
# b1         0 Bend            True 0.001     0
# drift_0    1 Drift           True     0     0
# q1         2 Quadrupole      True     0 0.002
# drift_1    3 Drift           True     0     0
# seq$end    4 Marker         False     0     0
# _end_point 4                False     0     0



import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=45.6e9, mass0=xt.ELECTRON_MASS_EV)

l1 = env.new_line(components=[
    env.new('b1', 'Bend', length=0.1)])
p0 = l1.build_particles(x=3e-3)
p = p0.copy()
l1.track(p)
xo.assert_allclose(p.px, 0, rtol=0, atol=1e-15)
env['b1'].knl[2] = 0.1
p = p0.copy()
l1.track(p)
assert np.abs(p.px[0]) > 1e-7

l2 = env.new_line(components=[
    env.new('q1', 'Quadrupole', length=0.1)])
p = p0.copy()
l2.track(p)
xo.assert_allclose(p.px, 0, rtol=0, atol=1e-15)
env['q1'].knl[2] = 0.1
p = p0.copy()
l2.track(p)
assert np.abs(p.px[0]) > 1e-7

l3 = env.new_line(components=[
    env.new('s1', 'Sextupole', length=0.1)])
p = p0.copy()
l3.track(p)
xo.assert_allclose(p.px, 0, rtol=0, atol=1e-15)
env['s1'].knl[2] = 0.1
p = p0.copy()
l3.track(p)
assert np.abs(p.px[0]) > 1e-7

l4 = env.new_line(components=[
    env.new('o1', 'Octupole', length=0.1)])
p = p0.copy()
l4.track(p)
xo.assert_allclose(p.px, 0, rtol=0, atol=1e-15)
env['o1'].knl[2] = 0.1
p = p0.copy()
l4.track(p)
assert np.abs(p.px[0]) > 1e-7

l5 = env.new_line(components=[
    env.new('rb1', 'RBend', length=0.1)])
p0 = l5.build_particles(x=3e-3)
p = p0.copy()
l5.track(p)
xo.assert_allclose(p.px, 0, rtol=0, atol=1e-15)
env['rb1'].knl[2] = 0.1
p = p0.copy()
l5.track(p)
assert np.abs(p.px[0]) > 1e-7



import xtrack as xt
import xobjects as xo
import numpy as np
import pathlib

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

def test_legacy_multiline_to_env_conversion():
    env = xt.Multiline.from_json(
        test_data_folder / 'hllhc15_thick/hllhc15_collider_thick_legacy_multiline.json')
    assert isinstance(env, xt.Environment)
    env.lhcb2.twiss_default['reverse'] = True

    assert env.lhcb1.element_dict is env.element_dict
    assert env.lhcb2.element_dict is env.element_dict

    assert 'drift_1' in env.lhcb1.element_names
    assert 'drift_1' not in env.lhcb2.element_names

    assert 'ip1' in env.lhcb1.element_names
    assert 'ip1' in env.lhcb2.element_names
    assert 'ip1/b2' not in env.lhcb2.element_names

    assert 'mqxfa.a1r1/lhcb1' in env.lhcb1.element_names
    assert 'mqxfa.a1r1/lhcb2' in env.lhcb2.element_names
    assert 'mqxfa.a1r1'  not in env.lhcb1.element_names
    assert 'mqxfa.a1r1'  not in env.lhcb2.element_names

    assert 'taxs.1l5/lhcb1' in env.lhcb1.element_names
    assert 'taxs.1l5/lhcb2' in env.lhcb2.element_names
    assert 'taxs.1l5'  not in env.lhcb1.element_names
    assert 'taxs.1l5'  not in env.lhcb2.element_names

    assert 'mq.12l6.b1' in env.lhcb1.element_names
    assert 'mq.12l6.b2' in env.lhcb2.element_names
    assert 'mq.12l6.b2/lhcb2' not in env.lhcb2.element_names

    assert 'drift_mcbyh.a4r1.b1..2' in env.lhcb1.element_names
    assert 'drift_mcbrdv.4l1.b2..2' in env.lhcb2.element_names

    twb1 = env.lhcb1.twiss4d()
    twb2 = env.lhcb2.twiss4d()

    assert np.all(twb1.rows['ip.*'].name ==
        np.array(['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8', 'ip1.l1']))
    assert np.all(twb2.rows['ip.*'].name ==
        np.array(['ip1', 'ip2', 'ip3', 'ip4', 'ip5', 'ip6', 'ip7', 'ip8', 'ip1.l1']))

    xo.assert_allclose(twb1.qx, 62.31, rtol=0, atol=1e-5)
    xo.assert_allclose(twb1.qy, 60.32, rtol=0, atol=1e-5)
    xo.assert_allclose(twb2.qx, 62.31, rtol=0, atol=1e-5)
    xo.assert_allclose(twb2.qy, 60.32, rtol=0, atol=1e-5)

    env['on_disp'] = 0
    env['on_x1'] = 10
    env['on_x5'] = 12

    twb1cross = env.lhcb1.twiss4d()
    twb2cross = env.lhcb2.twiss4d()

    xo.assert_allclose(twb1cross['px', 'ip1'], +10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twb2cross['px', 'ip1'], -10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twb1cross['py', 'ip5'], +12e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twb2cross['py', 'ip5'], -12e-6, rtol=0, atol=1e-9)

    # Check combined twiss
    twcross = env.twiss(method='4d')
    xo.assert_allclose(twcross.lhcb1['px', 'ip1'], +10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twcross.lhcb2['px', 'ip1'], -10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twcross.lhcb1['py', 'ip5'], +12e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twcross.lhcb2['py', 'ip5'], -12e-6, rtol=0, atol=1e-9)

    env.discard_trackers()
    assert env.lhcb1.tracker is None
    assert env.lhcb2.tracker is None

    mycontext = xo.ContextCpu()
    env.build_trackers(_context=mycontext)
    assert env.lhcb1.tracker is not None
    assert env.lhcb2.tracker is not None
    assert env.lhcb1.tracker._context is mycontext
    assert env.lhcb2.tracker._context is mycontext



import xtrack as xt

env = xt.Environment()

env.vars.default_to_zero = True

line = env.new_line(
    components=[
        env.new('c1', xt.Multipole, knl=['k0.1'], length=0.1, at=1.0),
        env.new('c2', xt.Multipole, knl=['k0.2'], length=0.1, at=2.0),
        env.new('obstacle', xt.LimitRect, min_x=-0.15, max_x=-0.05, at=2.5),
        env.new('c3', xt.Multipole, knl=['k0.3'], length=0.1, at=3.0),
        env.new('c4', xt.Multipole, knl=['k0.4'], length=0.1, at=4.0),
        env.new('end', xt.Marker, at=5.),
    ]
)
line.particle_ref = xt.Particles(p0c=10e9)
line.build_tracker()

line.get_table().show() # gives:
# Table: 10 rows, 11 cols
# name                   s element_type isthick isreplica parent_name ...
# drift_1                0 Drift           True     False None
# c1                     1 Multipole      False     False None
# drift_2                1 Drift           True     False None
# c2                     2 Multipole      False     False None
# obstacle               2 LimitRect      False     False None
# drift_3                2 Drift           True     False None
# c3                     3 Multipole      False     False None
# drift_4                3 Drift           True     False None
# c4                     4 Multipole      False     False None
# _end_point             4                False     False None

# Track a particle
p = line.particle_ref.copy()
line.track(p)

# Particle is lost at the obstacle
p.get_table().show(cols=['state', 'at_element'])
# is:
# particle_id           state at_element
# 0                         0          4

# Build a twin line for twiss
ltwiss = line.copy(shallow=True)

# Remove the obstacle
env.new('obstacle_marker', xt.Marker)
ltwiss.replace('obstacle', 'obstacle_marker')

# Match a bump to avoid the obstacle
opt = ltwiss.match(
    solve=False,
    betx=1.0, bety=1.0, x=0, px=0, y=0, py=0,
    targets=[
        xt.TargetSet(x=-0.1, px=0, at='obstacle_marker'),
        xt.TargetSet(x=0, px=0, at=xt.END),
    ],
    vary=xt.VaryList(['k0.1', 'k0.2', 'k0.3', 'k0.4'], step=1e-4)
)
opt.solve()

tw = ltwiss.twiss(betx=1.0, bety=1.0)
tw.plot('x')

# As all elements apart from the obstacle are in common, the bump is applied
# also in the original line

p = line.particle_ref.copy()
line.build_tracker()
line.track(p)

# The particle goes around the obstacle and is not lost:
p.get_table().show(cols=['state', 'at_element'])
# is:
# particle_id           state at_element
# 0                         1          0



import xtrack as xt

# Create an environment
env = xt.Environment()

# Create a line with two quadrupoles and a marker
line = env.new_line(name='myline', components=[
    env.new('q0', xt.Quadrupole, length=2.0, at=10.),
    env.new('q1', xt.Quadrupole, length=2.0, at=20.),
    env.new('m0', xt.Marker, at=40.),
    ])

tt0 = line.get_table()
tt0.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# _end_point            40            40            40

# Create a set of new elements to be placed
env.new('s1', xt.Sextupole, length=0.1, k2=0.2)
env.new('s2', xt.Sextupole, length=0.1, k2=-0.2)
env.new('m1', xt.Marker)
env.new('m2', xt.Marker)
env.new('m3', xt.Marker)

# Insert the new elements in the line
line.insert([
    env.place('s1', at=5.),
    env.place('s2', anchor='end', at=-5., from_='q1@start'),
    env.place(['m1', 'm2'], at='m0@start'),
    env.place('m3', at='m0@end'),
    ])

tt = line.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1..0             0         2.475          4.95
# s1                  4.95             5          5.05
# drift_1..2          5.05         7.025             9
# q0                     9            10            11
# drift_2..0            11         12.45          13.9
# s2                  13.9         13.95            14
# drift_2..2            14          16.5            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m1                    40            40            40
# m0                    40            40            40
# m3                    40            40            40



import xtrack as xt

env = xt.load_madx_lattice('../../test_data/lhc_2024/lhc.seq', reverse_lines=['lhcb2'])
env.vars.load_madx('../../test_data/lhc_2024/injection_optics.madx')

env.lhcb1.particle_ref = xt.Particles(p0c=7e12)
env.lhcb2.particle_ref = xt.Particles(p0c=7e12)

env.lhcb1.twiss4d().plot()
env.lhcb2.twiss4d(reverse=True).plot()





# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xobjects as xo
from .particles import Particles


_RecordIdentifier_getp_record_source = r'''
/*gpufun*/
/*gpuglmem*/ int8_t* RecordIdentifier_getp_record(RecordIdentifier record_id, LocalParticle* part){
    /*gpuglmem*/ int8_t* io_buffer = LocalParticle_get_io_buffer(part);
    if (io_buffer == NULL){
        return NULL;
    }

    int64_t buffer_id = RecordIdentifier_get_buffer_id(record_id);
    /*gpuglmem*/ int64_t* found_id = (/*gpuglmem*/ int64_t*)io_buffer;
    if (buffer_id != (*found_id)){
        printf("Error: buffer_id mismatch!\n");
        return NULL;
    }

    int64_t offset = RecordIdentifier_get_offset(record_id);

    return io_buffer + offset;
    }

'''

_RecordIndex_get_slot_source = r'''

/*gpufun*/
int64_t RecordIndex_get_slot(RecordIndex record_index){

    if (record_index == NULL){
        return -2;
    }
    int64_t capacity = RecordIndex_get_capacity(record_index);
    /*gpuglmem*/ uint32_t* num_recorded = RecordIndex_getp_num_recorded(record_index);

    if(*num_recorded >= capacity){
        return -1;
    }

    uint32_t slot;

    slot = atomic_add(num_recorded, 1);   //only_for_context opencl
    slot = atomicAdd(num_recorded, 1);    //only_for_context cuda

    #pragma omp atomic capture            //only_for_context cpu_openmp
    slot = (*num_recorded)++;             //only_for_context cpu_serial cpu_openmp

    if (slot >= capacity){
        *num_recorded = capacity;
        return -1;
    }
    return (int64_t)slot;
}
'''


class RecordIdentifier(xo.Struct):
    '''
    To be inserted in the beam element.
    '''
    buffer_id = xo.Int64
    offset = xo.Int64

    _extra_c_sources = [_RecordIdentifier_getp_record_source]
    _depends_on = [Particles._XoStruct]


class RecordIndex(xo.Struct):
    '''
    To be inserted in the record class.
    '''
    capacity = xo.Int64
    num_recorded = xo.UInt32
    _dummy = xo.UInt32 # to make sure the size is a multiple of 64 bits (not really needed)
    buffer_id = xo.Int64

    _extra_c_sources = [_RecordIndex_get_slot_source]


class IOBufferHeader(xo.Struct):
    buffer_id = xo.Int64

def new_io_buffer(_context=None, capacity=1048576):

    if _context is None:
        _context = xo.ContextCpu()

    iobuf = _context.new_buffer(capacity=capacity)
    head = IOBufferHeader(_buffer=iobuf)
    assert head._offset == 0
    assert head._buffer is iobuf
    head.buffer_id = np.random.randint(0, 2**60, dtype=np.int64)

    return iobuf

def init_internal_record(internal_record_class, capacity, io_buffer):

    init_dict = {}
    if np.isscalar(capacity):
        # One-level record
        capacity = int(capacity)
        for ff in internal_record_class._XoStruct._fields:
            if hasattr(ff.ftype, 'to_nplike'): #is array
                init_dict[ff.name] = capacity
    else:
        # Record with multiple subrecords
        init_dict = {}
        for ff in internal_record_class._XoStruct._fields:
            if ff.name in capacity.keys():
                subtable_class = ff.ftype
                init_dict[ff.name] = {}
                for sff in subtable_class._fields:
                    if hasattr(sff.ftype, 'to_nplike'): #is array
                        init_dict[ff.name][sff.name] = capacity[ff.name]

    record = internal_record_class(_buffer=io_buffer, **init_dict)

    if np.isscalar(capacity):
        # One-level record
        record._index.capacity = capacity
    else:
        # Record with multiple subrecords
        for kk in capacity.keys():
            getattr(record, kk)._index.capacity = capacity[kk]

    return record

def start_internal_logging(elements, record=None, io_buffer=None, capacity=None):

    assert io_buffer is not None, ('io_buffer must be provided.')

    if not isinstance (elements, (list, tuple)):
        elements = [elements]

    assert len(set([ee._internal_record_class for ee in elements])) == 1, (
        'All elements should have the same _interal_record_class.')

    if record is None:
        assert capacity is not None
        record = init_internal_record(elements[0]._internal_record_class, capacity,
                                      io_buffer)

    assert record._buffer is io_buffer, (
        "The record should be stored in the specified io_buffer")

    for ee in elements:
        ee._internal_record_id.offset = record._offset
        ee._internal_record_id.buffer_id = xo.Int64._from_buffer(
                                                record._buffer, 0)
        ee.io_buffer = io_buffer
        ee.record = record

    return record

def stop_internal_logging(elements=None):

    if not isinstance (elements, (list, tuple)):
        elements = [elements]

    for ee in elements:
        if hasattr(ee, 'io_buffer'):
            ee._internal_record_id.offset = 0
            ee._internal_record_id.buffer_id = 0
            ee.io_buffer = None

def start_internal_logging_for_elements_of_type(tracker, element_type, capacity):

    record = init_internal_record(element_type._internal_record_class, capacity,
                                  tracker.io_buffer)

    for nn in tracker.line.element_names:
        ee = tracker.line.element_dict[nn]

        if (hasattr(ee, 'io_buffer') and ee.io_buffer is not None
            and ee.io_buffer is not tracker.io_buffer):
            raise RuntimeError(f'The element `{nn}` has an io_buffer that is '
                'incompatible with the io_buffer of the tracker. Please clear '
                'the internal logging for the element using '
                '`stop_internal_record(element=...)`.')

        if isinstance(ee, element_type):
            start_internal_logging(ee, record, tracker.io_buffer)
    return record

def stop_internal_logging_for_elements_of_type(tracker, element_type):

    for ee in tracker.line.elements:
        if isinstance(ee, element_type):
            stop_internal_logging(ee)

def generate_get_record(ele_classname, record_classname):
    content = '''
/*gpufun*/
RECORDCLASSNAME ELECLASSNAME_getp_internal_record(ELECLASSNAME el, LocalParticle* part){
    RecordIdentifier record_id = ELECLASSNAME_getp__internal_record_id(el);
    if (RecordIdentifier_get_buffer_id(record_id) <= 0){
        return NULL;
    }
    else{
        return (RECORDCLASSNAME) RecordIdentifier_getp_record(record_id, part);
    }
    }
    '''.replace(
        'RECORDCLASSNAME', record_classname).replace('ELECLASSNAME', ele_classname)
    return content



"""
RF-Track <-> Xsuite interface

Authors: Andrea Latina, Gianni Iadarola
Date: 22.11.2023

"""

import numpy as np

class RFT_Element:
    """Beam element calling RF-Track.

    Xtrack will use RF-Track to track a bunch through an RF-Track element. This can
    include any RF-Track element, e.g., field maps, absorbers, travelling- and standing-
    wave structures, electron coolers and others.

    Parameters:
        - element: An RF-Track element.
    """

    # init xtrack variables
    iscollective = True # <-- To state that the element uses a python track method
    isthick = True

    def __init__(self, element):

        import RF_Track as rft
        self.length = element.get_length()
        self.lattice = rft.Lattice()
        self.lattice.append_ref(element)
        self.bunch_in = rft.Bunch6d()
        self.bunch_out = rft.Bunch6d()
        self.arr_for_rft = np.empty(0)
        self.arr_for_xt = np.empty(0)

    def track(self, particles, increment_at_element=False):

        p = particles
        if p.x.shape[0]==0:
           return

        # Prepare input for RF-Track
        self.p = 1. + p.delta
        self.pz = np.sqrt(self.p**2 - p.px**2 - p.py**2)
        self.arr_for_rft.resize((len(p.x), 10))
        self.arr_for_rft[:,0] = p.x * 1e3 # mm
        self.arr_for_rft[:,1] = p.px * 1e3 / self.pz # mrad
        self.arr_for_rft[:,2] = p.y * 1e3 # mm
        self.arr_for_rft[:,3] = p.py * 1e3 / self.pz # mrad
        self.arr_for_rft[:,5] = p.p0c * self.p / 1e6 # MeV
        self.arr_for_rft[:,6] = p.mass0 / 1e6 # MeV
        self.arr_for_rft[:,7] = p.q0
        self.arr_for_rft[:,8] = p.weight
        self.arr_for_rft[p.state==1,9] = np.nan # nan == not lost

        # Track the refernece particle
        import RF_Track as rft
        pref0 = rft.Bunch6d(np.array([0,0,0,0,0,p.p0c[0]/1e6,p.mass0/1e6,p.q0,p.weight.sum()]))
        pref1 = self.lattice.track(pref0)

        # Track the bunch
        self.bunch_in.set_phase_space(self.arr_for_rft)
        self.bunch_out = self.lattice.track(self.bunch_in)
        self.arr_for_xt = self.bunch_out.get_phase_space('%x %Px %y %Py %t %Pc %lost %id', 'all')

        # Update particles
        self.arr_for_xt = self.arr_for_xt[self.arr_for_xt[:,7].argsort()] # sort by particle id
        p.x  = self.arr_for_xt[:,0] / 1e3 # m
        p.px = self.arr_for_xt[:,1] * 1e6 / p.p0c # rad
        p.y  = self.arr_for_xt[:,2] / 1e3 # m
        p.py = self.arr_for_xt[:,3] * 1e6 / p.p0c # rad
        p.zeta += self.length - self.arr_for_xt[:,4] * p.beta0 / 1e3 # m
        p.delta = (self.arr_for_xt[:,5] * 1e6 - p.p0c) / p.p0c #
        p.state[np.isnan(self.arr_for_xt[:,6])!=np.isnan(self.arr_for_rft[:,9])] = -400 # lost in RF-Track



import xtrack as xt
import numpy as np
from xobjects.test_helpers import for_all_test_contexts

@for_all_test_contexts
def test_is_aperture(test_context):
    elements={
        'e0': xt.Bend(k0=0.4, h=0.3, length=1),
        'e1': xt.Replica(parent_name='e0'),
        'e2': xt.Replica(parent_name='e1'),
        'a0': xt.LimitRect(),
        'a1': xt.Replica(parent_name='a0'),
        'a2': xt.Replica(parent_name='a1'),
    }

    line = xt.Line(elements=elements, element_names=list(elements.keys()))
    line.build_tracker(_context=test_context)

    _is_aperture = xt.line._is_aperture

    assert not _is_aperture(line['e0'], line)
    assert not _is_aperture(line['e1'], line)
    assert not _is_aperture(line['e2'], line)
    assert _is_aperture(line['a0'], line)
    assert _is_aperture(line['a1'], line)
    assert _is_aperture(line['a2'], line)


@for_all_test_contexts
def test_has_backtrack(test_context):
    elements={
        'e0': xt.Bend(k0=0.4, h=0.3, length=1),
        'e1': xt.Replica(parent_name='e0'),
        'e2': xt.Replica(parent_name='e1'),
        'a0': xt.Drift(),
        'a1': xt.Replica(parent_name='a0'),
        'a2': xt.Replica(parent_name='a1'),
    }

    line = xt.Line(elements=elements, element_names=list(elements.keys()))
    line.build_tracker(_context=test_context)

    _has_backtrack = xt.line._has_backtrack
    _allow_loss_refinement = xt.line._allow_loss_refinement

    assert _has_backtrack(line['e0'], line)
    assert _has_backtrack(line['e1'], line)
    assert _has_backtrack(line['e2'], line)
    assert _has_backtrack(line['a0'], line)
    assert _has_backtrack(line['a1'], line)
    assert _has_backtrack(line['a2'], line)

    assert not _allow_loss_refinement(line['e0'], line)
    assert not _allow_loss_refinement(line['e1'], line)
    assert not _allow_loss_refinement(line['e2'], line)
    assert _allow_loss_refinement(line['a0'], line)
    assert _allow_loss_refinement(line['a1'], line)
    assert _allow_loss_refinement(line['a2'], line)

def test_is_drift_behaves_like_drift():

    bend = xt.Bend(k0=0.4, h=0.3, length=1)

    line = xt.Line(
        elements=[bend, xt.Replica(parent_name='e0'), xt.Replica(parent_name='e1'),
                            xt.Drift(length=1.), xt.Replica(parent_name='e3'),
                            xt.Replica(parent_name='e4')])
    length = line.get_length()

    tt = line.get_table()

    assert np.all(tt['name'][:-1] == ['e0', 'e1', 'e2', 'e3', 'e4', 'e5'])

    _is_drift = xt.line._is_drift
    _behaves_like_drift = xt.line._behaves_like_drift

    assert not _is_drift(line['e0'], line)
    assert not _is_drift(line['e1'], line)
    assert not _is_drift(line['e2'], line)
    assert _is_drift(line['e3'], line)
    assert _is_drift(line['e4'], line)
    assert _is_drift(line['e5'], line)

    assert not _behaves_like_drift(line['e0'], line)
    assert not _behaves_like_drift(line['e1'], line)
    assert not _behaves_like_drift(line['e2'], line)
    assert _behaves_like_drift(line['e3'], line)
    assert _behaves_like_drift(line['e4'], line)
    assert _behaves_like_drift(line['e5'], line)

    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(xt.Uniform(3), name='e0'),
            xt.Strategy(xt.Uniform(3, mode='thick'), name='e1'),
            xt.Strategy(xt.Uniform(3), name='e2')])
    assert_allclose = np.testing.assert_allclose
    assert_allclose(length, line.get_length(), rtol=0, atol=1e-15)

    assert np.all(line.get_table().name == [
        'e0_entry', 'e0..entry_map', 'drift_e0..0', 'e0..0', 'drift_e0..1',
        'e0..1', 'drift_e0..2', 'e0..2', 'drift_e0..3', 'e0..exit_map',
        'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0', 'e1..1', 'e1..2',
        'e1..exit_map', 'e1_exit', 'e2_entry', 'e2..entry_map',
        'drift_e2..0', 'e2..0', 'drift_e2..1', 'e2..1', 'drift_e2..2',
        'e2..2', 'drift_e2..3', 'e2..exit_map', 'e2_exit', 'e3', 'e4',
        'e5', '_end_point'])

    for nn, ee in line.items():
        if nn.startswith('drift'):
            assert _is_drift(ee, line)
            assert _behaves_like_drift(ee, line)
        elif nn in ['e3', 'e4', 'e5']:
            assert _is_drift(ee, line)
            assert _behaves_like_drift(ee, line)
        elif nn.endswith('_entry') or nn.endswith('_exit'): # Markers
            assert not _is_drift(ee, line)
            assert _behaves_like_drift(ee, line)
        else:
            assert not _is_drift(ee, line)
            assert not _behaves_like_drift(ee, line)

    line.cut_at_s([0.6, 1.5])
    assert_allclose(length, line.get_length(), rtol=0, atol=1e-15)
    assert_allclose(line.get_s_position('drift_e0..2..1'), 0.6, rtol=0, atol=1e-15)
    assert_allclose(line.get_s_position('e1..1..1'), 1.5, rtol=0, atol=1e-15)

    assert _is_drift(line['drift_e0..2..1'], line)
    assert _behaves_like_drift(line['drift_e0..2..1'], line)
    assert not _is_drift(line['e1..1..1'], line)
    assert not _behaves_like_drift(line['e1..1..1'], line)

    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(None),
            xt.Strategy(xt.Uniform(2), name='e3'),
            xt.Strategy(xt.Uniform(3, mode='thick'), name='e4'),
            xt.Strategy(xt.Uniform(4), name='e5')])

    assert np.all(line.get_table().name == [
        'e0_entry', 'e0..entry_map', 'drift_e0..0', 'e0..0', 'drift_e0..1',
        'e0..1', 'drift_e0..2..0', 'drift_e0..2..1', 'e0..2',
        'drift_e0..3', 'e0..exit_map', 'e0_exit', 'e1_entry',
        'e1..entry_map', 'e1..0', 'e1..1_entry', 'e1..1..0', 'e1..1..1',
        'e1..1_exit', 'e1..2', 'e1..exit_map', 'e1_exit', 'e2_entry',
        'e2..entry_map', 'drift_e2..0', 'e2..0', 'drift_e2..1', 'e2..1',
        'drift_e2..2', 'e2..2', 'drift_e2..3', 'e2..exit_map', 'e2_exit',
        'drift_e3..0', 'drift_e3..1', 'drift_e3..2', 'e4_entry', 'e4..0',
        'e4..1', 'e4..2', 'e4_exit', 'e5_entry', 'drift_e5..0',
        'drift_e5..1', 'drift_e5..2', 'drift_e5..3', 'drift_e5..4',
        'e5_exit', '_end_point'])

    for nn, ee in line.items():
        if nn.startswith('drift'):
            assert _is_drift(ee, line)
            assert _behaves_like_drift(ee, line)
        elif nn in ['e4..0', 'e4..1', 'e4..2']:
            assert _is_drift(ee, line)
            assert _behaves_like_drift(ee, line)
        elif nn.endswith('_entry') or nn.endswith('_exit'): # Markers
            assert not _is_drift(ee, line)
            assert _behaves_like_drift(ee, line)
        else:
            assert not _is_drift(ee, line)
            assert not _behaves_like_drift(ee, line)

def test_replica_loops():

    bend = xt.Bend(k0=0.4, h=0.3, length=1)

    elements = {
        # correct case
        'a0': bend,
        'a1': xt.Replica(parent_name='a0'),
        'a2': xt.Replica(parent_name='a1'),
        # create a simple loop
        'b0': xt.Replica(parent_name='b0'),
        # loop with two replicas
        'c0': xt.Replica(parent_name='c1'),
        'c1': xt.Replica(parent_name='c0'),
        # bigger loop
        'd0': xt.Replica(parent_name='d1'),
        'd1': xt.Replica(parent_name='d2'),
        'd2': xt.Replica(parent_name='d0'),
        'd3': xt.Replica(parent_name='d1'),
    }

    line = xt.Line(elements=elements, element_names=list(elements.keys()))

    ok = ['a1', 'a2']
    for kk in ok:
        assert line[kk].resolve(line) is line.get('a0')

    error = ['b0', 'c0', 'c1', 'd0', 'd1', 'd2', 'd3']
    for kk in error:
        try:
            line[kk].resolve(line)
        except RecursionError:
            pass
        else:
            raise Exception(f'Element {kk} should not be resolvable')


import contextlib
import gc
import pytest

import xobjects as xo


def assert_context_empty(context):
    """Assert that there are no active buffers in the given context.

    An active buffer is a buffer that has not been freed yet.
    """
    gc.collect()
    alive_buffer_count = len(context._buffers)
    if alive_buffer_count > 0:
        pytest.fail(f"There were {alive_buffer_count} active buffers after a "
                    f"test run, which points to a memory leak during the test "
                    f"session.")


@pytest.fixture(scope="function", autouse=True)
def cleanup(request):
    """Assert that there are no active buffers after each test run. If the test
    fails, then don't bother though.
    """
    tests_failed_before = request.session.testsfailed
    yield
    tests_failed_after = request.session.testsfailed
    test_did_fail = tests_failed_after > tests_failed_before
    if not test_did_fail:
        assert_context_empty(xo.context_default)


@pytest.fixture(scope="function")
def temp_context_default_func(mocker):
    """Temporarily set the default context to a new context in function scope.

    This is useful if a test uses a fixture that allocates buffers in the
    default context, and as a result the buffers are not freed after the test
    run. This context manager can be used to temporarily set the default
    context to a new context, which will have the effect of bypassing the
    default `cleanup` fixture.

    Unfortunately, due to the fact that pytest keeps the value yielded by a
    fixture in a cache longer than the duration of the test, we cannot
    assert_context_empty on the new temporary context here, as it will not
    yet be empty.
    """
    temp_context = xo.ContextCpu()
    mocker.patch.object(xo.typeutils, "context_default", temp_context)
    mocker.patch.object(xo, "context_default", temp_context)
    yield


@pytest.fixture(scope="module")
def temp_context_default_mod(module_mocker):
    """Module scope version of `temp_context_default_func` fixture."""
    module_mocker.patch.object(xo.typeutils, "context_default", xo.ContextCpu())
    yield



import xtrack as xt
import xobjects as xo
import numpy as np

magnet = xt.Magnet(k0=0.02, h=0., k1=0.01, length=2.,
                   k2=0.005, k3=0.03,
                   k1s=0.01, k2s=0.005, k3s=0.05,
                   knl=[0.003, 0.001, 0.01, 0.02, 4., 6e2, 7e6],
                   ksl=[-0.005, 0.002, -0.02, 0.03, -2, 700., 4e6])
magnet.integrator = 'yoshida4'
magnet.num_multipole_kicks = 100

p0 = xt.Particles(x=1e-2, y=2e-2, py=1e-3, delta=3e-2)

model_to_test = 'drift-kick-drift-exact'

m_ref = magnet.copy()
m_ref.model = 'bend-kick-bend'

m_uniform = magnet.copy()
m_uniform.model = model_to_test
m_uniform.integrator='uniform'
m_uniform.num_multipole_kicks = 50000

m_teapot = magnet.copy()
m_teapot.model = model_to_test
m_teapot.integrator='teapot'
m_teapot.num_multipole_kicks = 50000

m_yoshida = magnet.copy()
m_yoshida.model = model_to_test
m_yoshida.integrator='yoshida4'
m_yoshida.num_multipole_kicks = 100


p_ref = p0.copy()
p_uniform = p0.copy()
p_teapot = p0.copy()
p_yoshida = p0.copy()

m_ref.track(p_ref)
m_uniform.track(p_uniform)
m_teapot.track(p_teapot)
m_yoshida.track(p_yoshida)

xo.assert_allclose(p_ref.x, p_uniform.x, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.px, p_uniform.px, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.y, p_uniform.y, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.py, p_uniform.py, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.zeta, p_uniform.zeta, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.delta, p_uniform.delta, rtol=0, atol=5e-13)

xo.assert_allclose(p_ref.x, p_teapot.x, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.px, p_teapot.px, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.y, p_teapot.y, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.py, p_teapot.py, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.zeta, p_teapot.zeta, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.delta, p_teapot.delta, rtol=0, atol=5e-13)

xo.assert_allclose(p_ref.x, p_yoshida.x, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.px, p_yoshida.px, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.y, p_yoshida.y, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.py, p_yoshida.py, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.zeta, p_yoshida.zeta, rtol=0, atol=5e-13)
xo.assert_allclose(p_ref.delta, p_yoshida.delta, rtol=0, atol=5e-13)


import time

import xtrack as xt

from cpymad.madx import Madx

# Load the line
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(p0c=7e12, mass=xt.PROTON_MASS_EV)
collider = xt.Environment(lines={'lhcb1': line})
collider.build_trackers()

# mad model for refence
mad = Madx()
mad.call('../../test_data/hllhc15_noerrors_nobb/sequence.madx')
mad.use('lhcb1')

mad.input('''
savebeta, label=bir7b1, place=s.ds.l7.b1;
''')
tw_mad_ref = mad.twiss().dframe()

tw_ref = collider.lhcb1.twiss()

ele_start_match = 's.ds.l7.b1'
ele_end_match = 'e.ds.r7.b1'
tw_init = tw_ref.get_twiss_init(ele_start_match)

betx_end_match = tw_ref['betx', ele_end_match]
bety_end_match = tw_ref['bety', ele_end_match]
alfx_end_match = tw_ref['alfx', ele_end_match]
alfy_end_match = tw_ref['alfy', ele_end_match]
dx_end_match = tw_ref['dx', ele_end_match]
dpx_end_match = tw_ref['dpx', ele_end_match]
mux_end_match = tw_ref['mux', ele_end_match]
muy_end_match = tw_ref['muy', ele_end_match]

betx_at_ip7 = tw_ref['betx', 'ip7']
bety_at_ip7 = tw_ref['bety', 'ip7']
alfx_at_ip7 = tw_ref['alfx', 'ip7']
alfy_at_ip7 = tw_ref['alfy', 'ip7']
dx_at_ip7 = tw_ref['dx', 'ip7']
dpx_at_ip7 = tw_ref['dpx', 'ip7']

scale = 23348.89927
scmin = 0.03*7000./line.vars['nrj']._value
qtlimitx28 = 1.0*225.0/scale
qtlimitx15 = 1.0*205.0/scale
qtlimit2 = 1.0*160.0/scale
qtlimit3 = 1.0*200.0/scale
qtlimit4 = 1.0*125.0/scale
qtlimit5 = 1.0*120.0/scale
qtlimit6 = 1.0*90.0/scale


# Break something

collider.vars['kqt13.l7b1'] = collider.vars['kqt13.l7b1']._value * 1.1
collider.vars['kqt12.l7b1'] = collider.vars['kqt12.l7b1']._value * 1.1
collider.vars['kqtl11.l7b1'] = collider.vars['kqtl11.l7b1']._value * 0.9
collider.vars['kqtl10.l7b1'] = collider.vars['kqtl10.l7b1']._value * 1.1
collider.vars['kqtl9.l7b1'] = collider.vars['kqtl9.l7b1']._value * 0.8
collider.vars['kqtl8.l7b1'] = collider.vars['kqtl8.l7b1']._value * 0.9
collider.vars['kqtl7.l7b1'] = collider.vars['kqtl7.l7b1']._value * 1.12
collider.vars['kq6.l7b1'] = collider.vars['kq6.l7b1']._value * 0.9
collider.vars['kq6.r7b1'] = collider.vars['kq6.r7b1']._value * 1.1
collider.vars['kqtl7.r7b1'] = collider.vars['kqtl7.r7b1']._value * 1.1
collider.vars['kqtl8.r7b1'] = collider.vars['kqtl8.r7b1']._value * 0.85
collider.vars['kqtl9.r7b1'] = collider.vars['kqtl9.r7b1']._value * 0.95
collider.vars['kqtl10.r7b1'] = collider.vars['kqtl10.r7b1']._value * 0.85
collider.vars['kqtl11.r7b1'] = collider.vars['kqtl11.r7b1']._value * 1.15
collider.vars['kqt12.r7b1'] = collider.vars['kqt12.r7b1']._value * 0.9
collider.vars['kqt13.r7b1'] = collider.vars['kqt13.r7b1']._value * 1.12


mad.use('lhcb1')
# Same in mad
mad_pertub = mad.input(f'''
kqt13.l7b1  = {collider.vars['kqt13.l7b1']._value};
kqt12.l7b1  = {collider.vars['kqt12.l7b1']._value};
kqtl11.l7b1 = {collider.vars['kqtl11.l7b1']._value};
kqtl10.l7b1 = {collider.vars['kqtl10.l7b1']._value};
kqtl9.l7b1  = {collider.vars['kqtl9.l7b1']._value};
kqtl8.l7b1  = {collider.vars['kqtl8.l7b1']._value};
kqtl7.l7b1  = {collider.vars['kqtl7.l7b1']._value};
kq6.l7b1    = {collider.vars['kq6.l7b1']._value};
kq6.r7b1    = {collider.vars['kq6.r7b1']._value};
kqtl7.r7b1  = {collider.vars['kqtl7.r7b1']._value};
kqtl8.r7b1  = {collider.vars['kqtl8.r7b1']._value};
kqtl9.r7b1  = {collider.vars['kqtl9.r7b1']._value};
kqtl10.r7b1 = {collider.vars['kqtl10.r7b1']._value};
kqtl11.r7b1 = {collider.vars['kqtl11.r7b1']._value};
kqt12.r7b1  = {collider.vars['kqt12.r7b1']._value};
kqt13/r7b1  = {collider.vars['kqt13.r7b1']._value};
''')

tw_before = collider.lhcb1.twiss()
tw_mad_before = mad.twiss().dframe()

t_start = time.perf_counter()
collider.match(
    #verbose=True,
    start=ele_start_match,
    end=ele_end_match,
    init=tw_init,
    targets=[
        xt.Target(line='lhcb1', at='ip7',        tar='dx',   value=dx_at_ip7, tol=1e-3),
        xt.Target(line='lhcb1', at='ip7',        tar='dpx',  value=dpx_at_ip7, tol=1e-5),
        xt.Target(line='lhcb1', at='ip7',        tar='betx', value=betx_at_ip7, tol=1e-3),
        xt.Target(line='lhcb1', at='ip7',        tar='bety', value=bety_at_ip7, tol=1e-3),
        xt.Target(line='lhcb1', at='ip7',        tar='alfx', value=alfx_at_ip7, tol=1e-5),
        xt.Target(line='lhcb1', at='ip7',        tar='alfy', value=alfy_at_ip7, tol=1e-5),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='alfx', value=alfx_end_match, tol=1e-5),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='alfy', value=alfy_end_match, tol=1e-5),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='betx', value=betx_end_match, tol=1e-3),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='bety', value=bety_end_match, tol=1e-3),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='dx',   value=dx_end_match, tol=1e-3),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='dpx',  value=dpx_end_match, tol=1e-5),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='mux',  value=mux_end_match, tol=1e-5),
        xt.Target(line='lhcb1', at='e.ds.r7.b1', tar='muy', value=muy_end_match, tol=1e-5),
        # xt.TargetInequality('bety', '<', 180.49-0.3, line='lhcb1', at='mq.11l7.b1'),
        # xt.TargetInequality('bety', '<', 174.5,      line='lhcb1', at='mq.9l7.b1'),
        # xt.TargetInequality('bety', '<', 176.92,     line='lhcb1', at='mq.8r7.b1'),
        # xt.TargetInequality('bety', '<', 179,        line='lhcb1', at='mq.10r7.b1'),
    ],
    vary=[
        xt.Vary('kqt13.l7b1',  step=1.0E-9, limits=(-qtlimit5, qtlimit5)),
        xt.Vary('kqt12.l7b1',  step=1.0E-9, limits=(-qtlimit5, qtlimit5)),
        xt.Vary('kqtl11.l7b1', step=1.0E-9, limits=(-qtlimit4*300./550., qtlimit4*300./550.)),
        xt.Vary('kqtl10.l7b1', step=1.0E-9, limits=(-qtlimit4*500./550., qtlimit4*500./550.)),
        xt.Vary('kqtl9.l7b1',  step=1.0E-9, limits=(-qtlimit4*400./550., qtlimit4*400./550.)),
        xt.Vary('kqtl8.l7b1',  step=1.0E-9, limits=(-qtlimit4*300./550., qtlimit4*300./550.)),
        xt.Vary('kqtl7.l7b1',  step=1.0E-9, limits=(-qtlimit4, qtlimit4)),
        xt.Vary('kq6.l7b1',    step=1.0E-9, limits=(-qtlimit6, qtlimit6)),
        xt.Vary('kq6.r7b1',    step=1.0E-9, limits=(-qtlimit6, qtlimit6)),
        xt.Vary('kqtl7.r7b1',  step=1.0E-9, limits=(-qtlimit4, qtlimit4)),
        xt.Vary('kqtl8.r7b1',  step=1.0E-9, limits=(-qtlimit4*550./550., qtlimit4*550./550.)),
        xt.Vary('kqtl9.r7b1',  step=1.0E-9, limits=(-qtlimit4*500./550., qtlimit4*500./550.)),
        xt.Vary('kqtl10.r7b1', step=1.0E-9, limits=(-qtlimit4, qtlimit4)),
        xt.Vary('kqtl11.r7b1', step=1.0E-9, limits=(-qtlimit4, qtlimit4)),
        xt.Vary('kqt12.r7b1',  step=1.0E-9, limits=(-qtlimit5, qtlimit5)),
        xt.Vary('kqt13.r7b1',  step=1.0E-9, limits=(-qtlimit5, qtlimit5)),
    ]
)
t_end = time.perf_counter()


tw_after = collider.lhcb1.twiss()

t1 = time.perf_counter()
mad.input(f'''

qtlimitx28 = {qtlimitx28};
qtlimitx15 = {qtlimitx15};
qtlimit2 = {qtlimit2};
qtlimit3 = {qtlimit3};
qtlimit4 =  {qtlimit4};
qtlimit5 =  {qtlimit5};
qtlimit6 =  {qtlimit6};

use,sequence=lhcb1,range=s.ds.l7.b1/e.ds.r7.b1;
match,      sequence=lhcb1, beta0=bir7b1;
weight,mux=10,muy=10;
constraint, sequence=lhcb1, range=ip7,dx={dx_at_ip7},dpx ={dpx_at_ip7};
constraint, sequence=lhcb1, range=ip7,betx={betx_at_ip7},bety={bety_at_ip7};
constraint, sequence=lhcb1, range=ip7,alfx={alfx_at_ip7},alfy={alfy_at_ip7};
constraint, sequence=lhcb1, range=e.ds.r7.b1,alfx={alfx_end_match},alfy={alfy_end_match};
constraint, sequence=lhcb1, range=e.ds.r7.b1,betx={betx_end_match},bety={bety_end_match};
constraint, sequence=lhcb1, range=e.ds.r7.b1,dx={dx_end_match},dpx={dpx_end_match};
constraint, sequence=lhcb1, range=e.ds.r7.b1,   mux={mux_end_match};
constraint, sequence=lhcb1, range=e.ds.r7.b1,   muy={muy_end_match};
vary, name=kqt13.l7b1,  step=1.0E-9, lower=-qtlimit5, upper=qtlimit5;
vary, name=kqt12.l7b1,  step=1.0E-9, lower=-qtlimit5, upper=qtlimit5;
vary, name=kqtl11.l7b1, step=1.0E-9, lower=-qtlimit4*300./550., upper=qtlimit4*300./550.;
vary, name=kqtl10.l7b1, step=1.0E-9, lower=-qtlimit4*500./550., upper=qtlimit4*500./550.;
vary, name=kqtl9.l7b1,  step=1.0E-9, lower=-qtlimit4*400./550., upper=qtlimit4*400./550.;
vary, name=kqtl8.l7b1,  step=1.0E-9, lower=-qtlimit4*300./550., upper=qtlimit4*300./550.;
vary, name=kqtl7.l7b1,  step=1.0E-9, lower=-qtlimit4, upper=qtlimit4;
vary, name=kq6.l7b1,    step=1.0E-9, lower=-qtlimit6, upper=qtlimit6;
vary, name=kq6.r7b1,    step=1.0E-9, lower=-qtlimit6, upper=qtlimit6;
vary, name=kqtl7.r7b1,  step=1.0E-9, lower=-qtlimit4, upper=qtlimit4;
vary, name=kqtl8.r7b1,  step=1.0E-9, lower=-qtlimit4*550./550., upper=qtlimit4*550./550.;
vary, name=kqtl9.r7b1,  step=1.0E-9, lower=-qtlimit4*500./550., upper=qtlimit4*500./550.;
vary, name=kqtl10.r7b1, step=1.0E-9, lower=-qtlimit4, upper=qtlimit4;
vary, name=kqtl11.r7b1, step=1.0E-9, lower=-qtlimit4, upper=qtlimit4;
vary, name=kqt12.r7b1,  step=1.0E-9, lower=-qtlimit5, upper=qtlimit5;
vary, name=kqt13.r7b1,  step=1.0E-9, lower=-qtlimit5, upper=qtlimit5;
jacobian,calls=15, tolerance=1e-20, bisec=3;
endmatch;
''')
t2 = time.perf_counter()
mad.use(sequence='lhcb1')
tw_mad_after = mad.twiss().dframe()
print(f"Xsuite matching time: {t_end - t_start:0.4f} seconds")
print(f"MAD-X matching time:  {t2 - t1:0.4f} seconds")





import time

import xtrack as xt

from cpymad.madx import Madx

# xt._print.suppress = True

# Load the line
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(p0c=7e12, mass=xt.PROTON_MASS_EV)
collider = xt.Environment(lines={'lhcb1': line})
collider.build_trackers()

scale = 23348.89927
scmin = 0.03*7000./line.vars['nrj']._value
qtlimitx28 = 1.0*225.0/scale
qtlimitx15 = 1.0*205.0/scale
qtlimit2 = 1.0*160.0/scale
qtlimit3 = 1.0*200.0/scale
qtlimit4 = 1.0*125.0/scale
qtlimit5 = 1.0*120.0/scale
qtlimit6 = 1.0*90.0/scale

collider.vars.vary_default.update({
    'kqt13.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
    'kqt12.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
    'kqtl11.l7b1': {'step': 1.0E-9, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
    'kqtl10.l7b1': {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
    'kqtl9.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*400./550., qtlimit4*400./550.)},
    'kqtl8.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*300./550., qtlimit4*300./550.)},
    'kqtl7.l7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
    'kq6.l7b1':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
    'kq6.r7b1':    {'step': 1.0E-9, 'limits': (-qtlimit6, qtlimit6)},
    'kqtl7.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
    'kqtl8.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*550./550., qtlimit4*550./550.)},
    'kqtl9.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit4*500./550., qtlimit4*500./550.)},
    'kqtl10.r7b1': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
    'kqtl11.r7b1': {'step': 1.0E-9, 'limits': (-qtlimit4, qtlimit4)},
    'kqt12.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
    'kqt13.r7b1':  {'step': 1.0E-9, 'limits': (-qtlimit5, qtlimit5)},
})
tw_ref = collider.lhcb1.twiss()

ele_start_match = 's.ds.l7.b1'
ele_end_match = 'e.ds.r7.b1'
tw_init = tw_ref.get_twiss_init(ele_start_match)

betx_end_match = tw_ref['betx', ele_end_match]
bety_end_match = tw_ref['bety', ele_end_match]
alfx_end_match = tw_ref['alfx', ele_end_match]
alfy_end_match = tw_ref['alfy', ele_end_match]
dx_end_match = tw_ref['dx', ele_end_match]
dpx_end_match = tw_ref['dpx', ele_end_match]
mux_end_match = tw_ref['mux', ele_end_match]
muy_end_match = tw_ref['muy', ele_end_match]

betx_at_ip7 = tw_ref['betx', 'ip7']
bety_at_ip7 = tw_ref['bety', 'ip7']
alfx_at_ip7 = tw_ref['alfx', 'ip7']
alfy_at_ip7 = tw_ref['alfy', 'ip7']
dx_at_ip7 = tw_ref['dx', 'ip7']
dpx_at_ip7 = tw_ref['dpx', 'ip7']


# Break something

perturbed_vars = {}
perturbed_vars['kqt13.l7b1'] = collider.vars['kqt13.l7b1']._value * 1.1
perturbed_vars['kqt12.l7b1'] = collider.vars['kqt12.l7b1']._value * 1.1
perturbed_vars['kqtl11.l7b1'] = collider.vars['kqtl11.l7b1']._value * 0.9
perturbed_vars['kqtl10.l7b1'] = collider.vars['kqtl10.l7b1']._value * 1.1
perturbed_vars['kqtl9.l7b1'] = collider.vars['kqtl9.l7b1']._value * 0.8
perturbed_vars['kqtl8.l7b1'] = collider.vars['kqtl8.l7b1']._value * 0.9
perturbed_vars['kqtl7.l7b1'] = collider.vars['kqtl7.l7b1']._value * 1.12
perturbed_vars['kq6.l7b1'] = collider.vars['kq6.l7b1']._value * 0.9
perturbed_vars['kq6.r7b1'] = collider.vars['kq6.r7b1']._value * 1.1
perturbed_vars['kqtl7.r7b1'] = collider.vars['kqtl7.r7b1']._value * 1.1
perturbed_vars['kqtl8.r7b1'] = collider.vars['kqtl8.r7b1']._value * 0.85
perturbed_vars['kqtl9.r7b1'] = collider.vars['kqtl9.r7b1']._value * 0.95
perturbed_vars['kqtl10.r7b1'] = collider.vars['kqtl10.r7b1']._value * 0.85
perturbed_vars['kqtl11.r7b1'] = collider.vars['kqtl11.r7b1']._value * 1.15
perturbed_vars['kqt12.r7b1'] = collider.vars['kqt12.r7b1']._value * 0.9
perturbed_vars['kqt13.r7b1'] = collider.vars['kqt13.r7b1']._value * 1.12

tw_before = collider.lhcb1.twiss()

for i_repeat in range(1):

    for nn, vv in perturbed_vars.items():
        collider.vars[nn]= vv

    t_start = time.perf_counter()
    tw0 = collider.twiss()
    opt = collider.match(
        solve=False,
        start=ele_start_match,
        end=ele_end_match,
        init=tw_ref, init_at=xt.START,
        targets=[
            xt.TargetList(
                ('betx', 'bety', 'alfx', 'alfy', 'dx', 'dpx'),
                at='ip7', line='lhcb1', value=tw_ref),
            xt.TargetList(
                ('alfx', 'alfy', 'betx', 'bety', 'dx', 'dpx', 'mux', 'muy'),
                at='e.ds.r7.b1',line='lhcb1', value=tw_ref),
            ],
        vary=[
            xt.VaryList(
                ('kqt13.l7b1', 'kqt12.l7b1', 'kqtl11.l7b1', 'kqtl10.l7b1',
                 'kqtl9.l7b1', 'kqtl8.l7b1', 'kqtl7.l7b1', 'kq6.l7b1',
                 'kq6.r7b1', 'kqtl7.r7b1', 'kqtl8.r7b1', 'kqtl9.r7b1',
                 'kqtl10.r7b1', 'kqtl11.r7b1', 'kqt12.r7b1', 'kqt13.r7b1'))
        ]
    )

    match_res = opt.solve()
    t_end = time.perf_counter()
    print(f"Matching time: {t_end - t_start:0.4f} seconds")


tw_after = collider.lhcb1.twiss()


_err = opt._err
x_final = opt.solver._xbest

n_repeat_err_call = 100
t0 = time.perf_counter()
for _ in range(n_repeat_err_call):
    _err(x_final)
t1 = time.perf_counter()
print(f"Error call time: {1000 * (t1 - t0)/n_repeat_err_call:0.4f} ms")

n_repeat_set_knobs = 100
t0 = time.perf_counter()
for _ in range(n_repeat_set_knobs):
    knob_values = _err._x_to_knobs(x_final)
    for kk, vv in zip(_err.vary, knob_values):
        _err.actions[0].line.vars[kk.name] = vv
t1 = time.perf_counter()
print(f"Set knobs time: {1000 * (t1 - t0)/n_repeat_set_knobs:0.4f} ms")

last_data = _err._last_data
n_repeat_eval_targets = 100
t0 = time.perf_counter()
for _ in range(n_repeat_eval_targets):
    for tt in _err.targets:
        tt.eval(last_data)
t1 = time.perf_counter()
print(f"Evaluate targets time: {1000 * (t1 - t0)/n_repeat_eval_targets:0.4f} ms")


tw_init = tw_ref.get_twiss_init(ele_start_match)
ele_index_start = line.element_names.index(ele_start_match)
ele_index_end = line.element_names.index(ele_end_match)

ttt = collider.twiss(
        start=[ele_index_start],
        end=[ele_index_end],
        init=tw_init,
        _keep_initial_particles=True,
        _keep_tracking_data=True,
        )

n_repeat_twiss = 100
t0 = time.perf_counter()
for _ in range(n_repeat_twiss):
    collider.twiss(
        start=[ele_index_start],
        end=[ele_index_end],
        init=tw_init,
        _ebe_monitor=[ttt.lhcb1.tracking_data],
        _initial_particles=[ttt.lhcb1._initial_particles]
        )
t1 = time.perf_counter()
print(f"Twiss time: {1000 * (t1 - t0)/n_repeat_twiss:0.4f} ms")

n_repeat_tracking = 100
p_test = [ttt.lhcb1._initial_particles.copy() for _ in range(n_repeat_tracking)]
t0 = time.perf_counter()
for ii in range(n_repeat_tracking):
    collider.lhcb1.track(
        p_test[ii],
        start=ele_index_start,
        end=ele_index_end,
        turn_by_turn_monitor=ttt.lhcb1.tracking_data
        )
t1 = time.perf_counter()
print(f"Tracking time: {1000 * (t1 - t0)/n_repeat_tracking:0.4f} ms")



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_thick/hllhc15_collider_thick.json')
collider.build_trackers()

collider.lhcb1.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['reverse'] = True

collider.vars['on_x1vs'] = 123
collider.vars['on_sep1h'] = 2
collider.vars['on_x1vs'] = 200
collider.vars['on_sep1v'] = -3

# tw = collider.lhcb1.twiss()                  # Fails on closed orbit search
tw = collider.lhcb1.twiss(co_search_at='ip7') # Successful closed orbit search

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(tw.s, tw.x)
plt.plot(tw.s, tw.y)
plt.show()


import pathlib

from cpymad.madx import Madx

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

import numpy as np

@for_all_test_contexts
def test_chromatic_functions_vs_madx(test_context):

    collider = xt.Environment.from_json(test_data_folder /
                                'hllhc15_thick/hllhc15_collider_thick.json')
    collider['lhcb1'].twiss_default['method'] = '4d'
    collider['lhcb2'].twiss_default['method'] = '4d'


    # Slice to have the exit markers
    collider.lhcb1.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(1, mode='thick'))])
    collider.lhcb2.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(1, mode='thick'))])

    collider.build_trackers(_context=test_context)

    mad = Madx(stdout=False)
    mad.input(f"""
    call,file="{str(test_data_folder)}/hllhc15_thick/lhc.seq";
    call,file="{str(test_data_folder)}/hllhc15_thick/hllhc_sequence.madx";
    beam, sequence=lhcb1, particle=proton, pc=7000;
    beam, sequence=lhcb2, particle=proton, pc=7000, bv=-1;
    call,file="{str(test_data_folder)}/hllhc15_thick/opt_round_150_1500.madx";
    """)

    for line_name in ['lhcb1', 'lhcb2']:

        if line_name == 'lhcb1':
            line = collider.lhcb1
            mad.use(sequence="lhcb1")
        elif line_name == 'lhcb2':
            line = collider.lhcb2
            mad.use(sequence="lhcb2")
            line.twiss_default['reverse'] = True

        tw = line.twiss()
        twmad = mad.twiss(chrom=True)

        tw_test = tw.rows['.*_exit']

        wx_ref = np.interp(tw_test['s'], twmad['s'], twmad['wx'])
        wy_ref = np.interp(tw_test['s'], twmad['s'], twmad['wy'])
        phix_ref = np.interp(tw_test['s'], twmad['s'], twmad['phix'])
        phiy_ref = np.interp(tw_test['s'], twmad['s'], twmad['phiy'])

        zx_ref = wx_ref * np.exp(1j * 2 * np.pi * phix_ref)
        zy_ref = wy_ref * np.exp(1j * 2 * np.pi * phiy_ref)
        ax_ref = np.imag(zx_ref)
        ay_ref = np.imag(zy_ref)
        bx_ref = np.real(zx_ref)
        by_ref = np.real(zy_ref)

        xo.assert_allclose(tw_test.wx_chrom, wx_ref, rtol=0, atol=2e-3 * np.max(wx_ref))
        xo.assert_allclose(tw_test.wy_chrom, wy_ref, rtol=0, atol=2e-3 * np.max(wy_ref))
        xo.assert_allclose(tw_test.ax_chrom, ax_ref, rtol=0, atol=2e-3 * np.max(ax_ref))
        xo.assert_allclose(tw_test.ay_chrom, ay_ref, rtol=0, atol=2e-3 * np.max(ay_ref))
        xo.assert_allclose(tw_test.bx_chrom, bx_ref, rtol=0, atol=2e-3 * np.max(bx_ref))
        xo.assert_allclose(tw_test.by_chrom, by_ref, rtol=0, atol=2e-3 * np.max(by_ref))

        # Open twiss
        init = tw.get_twiss_init('ip3')
        tw_open = line.twiss(start='ip3', end='ip6', init=init,
                            compute_chromatic_properties=True)

        tw_ref_open = tw.rows['ip3':'ip6']
        xo.assert_allclose(tw_open.wx_chrom[:-1], tw_ref_open.wx_chrom,
                        rtol=0, atol=2e-3 * np.max(tw_ref_open.wx_chrom))
        xo.assert_allclose(tw_open.wy_chrom[:-1], tw_ref_open.wy_chrom,
                        rtol=0, atol=2e-3 * np.max(tw_ref_open.wy_chrom))
        xo.assert_allclose(tw_open.ax_chrom[:-1], tw_ref_open.ax_chrom,
                            rtol=0, atol=2e-3 * np.max(tw_ref_open.ax_chrom))
        xo.assert_allclose(tw_open.ay_chrom[:-1], tw_ref_open.ay_chrom,
                                rtol=0, atol=2e-3 * np.max(tw_ref_open.ay_chrom))
        xo.assert_allclose(tw_open.bx_chrom[:-1], tw_ref_open.bx_chrom,
                                rtol=0, atol=2e-3 * np.max(tw_ref_open.bx_chrom))


import time

import xtrack as xt

import xtrack._temp.lhc_match as lm

default_tol = {None: 1e-8, 'betx': 1e-6, 'bety': 1e-6} # to have no rematching w.r.t. madx

collider = xt.Environment.from_json(
    "../../test_data/hllhc15_thick/hllhc15_collider_thick.json")
collider.build_trackers()
collider.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

c0 = collider.copy()

h_correctors_ip1_b1 = ['acbh16.r8b1', 'acbh14.l1b1', 'acbh12.l1b1',
                       'acbh13.r1b1', 'acbh15.r1b1', 'acbh15.l2b1']
v_correctors_ip1_b1 = ['acbv15.r8b1', 'acbv15.l1b1', 'acbv13.l1b1',
                       'acbv12.r1b1', 'acbv14.r1b1', 'acbv16.l2b1']
h_correctors_ip5_b1 = ['acbh16.r4b1', 'acbh14.l5b1', 'acbh12.l5b1',
                       'acbh13.r5b1', 'acbh15.r5b1', 'acbh15.l6b1']
v_correctors_ip5_b1 = ['acbv15.r4b1', 'acbv15.l5b1', 'acbv13.l5b1',
                       'acbv12.r5b1', 'acbv14.r5b1', 'acbv16.l6b1']

h_correctors_ip1_b2 = ['acbh15.r8b2', 'acbh15.l1b2', 'acbh13.l1b2',
                       'acbh12.r1b2', 'acbh14.r1b2', 'acbh16.l2b2']
v_correctors_ip1_b2 = ['acbv16.r8b2', 'acbv14.l1b2', 'acbv12.l1b2',
                       'acbv13.r1b2', 'acbv15.r1b2', 'acbv15.l2b2']
h_correctors_ip5_b2 = ['acbh12.r5b2', 'acbh14.r5b2', 'acbh16.l6b2',
                       'acbh15.r4b2', 'acbh15.l5b2', 'acbh13.l5b2']
v_correctors_ip5_b2 = ['acbv16.r4b2', 'acbv14.l5b2', 'acbv12.l5b2',
                       'acbv13.r5b2', 'acbv15.r5b2', 'acbv15.l6b2']

correctors = {}
correctors['ip1_x_b1'] = h_correctors_ip1_b1
correctors['ip1_y_b1'] = v_correctors_ip1_b1
correctors['ip5_x_b1'] = h_correctors_ip5_b1
correctors['ip5_y_b1'] = v_correctors_ip5_b1
correctors['ip1_x_b2'] = h_correctors_ip1_b2
correctors['ip1_y_b2'] = v_correctors_ip1_b2
correctors['ip5_x_b2'] = h_correctors_ip5_b2
correctors['ip5_y_b2'] = v_correctors_ip5_b2

# Wipe corrector strengths
for kk in correctors:
    for cc in correctors[kk]:
        collider.vars[cc] = 0

acb_limits = (-800.e-6, 800e-6)

knobs_to_compensate = {
    'on_x1hl': dict(value=295, ip='ip1', plane='x'),
    'on_x1hs': dict(value=295, ip='ip1', plane='y'),
    'on_x1vl': dict(value=295, ip='ip1', plane='y'),
    'on_x1vs': dict(value=295, ip='ip1', plane='x'),
    'on_x5hl': dict(value=295, ip='ip5', plane='x'),
    'on_x5hs': dict(value=295, ip='ip5', plane='y'),
    'on_x5vl': dict(value=295, ip='ip5', plane='y'),
    'on_x5vs': dict(value=295, ip='ip5', plane='x'),
    'on_sep1h': dict(value=1, ip='ip1', plane='x'),
    'on_sep1v': dict(value=1, ip='ip1', plane='y'),
    'on_sep5h': dict(value=1, ip='ip5', plane='x'),
    'on_sep5v': dict(value=1, ip='ip5', plane='y'),
}

start_ele_before_cycle = []
for line_name in ['lhcb1', 'lhcb2']:
    line = collider[line_name]
    start_ele_before_cycle.append(line.element_names[0])
    line.cycle('ip3', inplace=True)

for kk in knobs_to_compensate:

    disp_knob_name = kk.replace('on_', 'on_d')

    plane = knobs_to_compensate[kk]['plane']
    ip = knobs_to_compensate[kk]['ip']
    ref_val = knobs_to_compensate[kk]['value']

    for line_name in ['lhcb1', 'lhcb2']:
        line = collider[line_name]
        beam_name = line_name[-2:]

        left_ip = {'ip1': 'ip8', 'ip5': 'ip4'}[ip]
        right_ip = {'ip1': 'ip2', 'ip5': 'ip6'}[ip]

        tw_ref = line.twiss()
        opt = line.match_knob(
            run=False,
            knob_name=disp_knob_name,
            knob_value_start=0, knob_value_end=ref_val,
            start=left_ip, end=right_ip,
            init=tw_ref, init_at=xt.START,
            vary=[
                xt.VaryList(correctors[f'{ip}_{plane}_{beam_name}'],
                            step=1e-10, limits=acb_limits),
                ],
            targets=[
                # Constraints on dispersion
                xt.Target('d' + plane, value=tw_ref, at=ip, tol=1e-6),
                xt.Target('d' + plane, value=tw_ref, at=right_ip, tol=1e-6),
                # Constraints on orbit
                xt.TargetSet([plane, 'p' + plane], value=tw_ref, at=f'e.ds.l{ip[-1:]}.{beam_name}'),
                xt.TargetSet([plane, 'p' + plane], value=tw_ref, at=f'e.ds.l{right_ip[-1:]}.{beam_name}'),
            ],
        )

        temp_expr = line.vars[kk]._expr
        line.vars[kk] = ref_val
        opt.solve()
        line.vars[kk] = temp_expr
        opt.generate_knob()

# cycle back
for line_name, estart  in zip(['lhcb1', 'lhcb2'], start_ele_before_cycle):
    line = collider[line_name]
    line.cycle(estart, inplace=True)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# Periodic twiss of the full ring
tw_p = line.twiss()

# Periodic twiss of an arc cell
tw = line.twiss(method='4d', start='mq.14r6.b1', end='mq.16r6.b1', init='periodic')

#!end-doc-part

# Plot

# Choose the twiss to plot

import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1)
spbet = plt.subplot(2,1,1)
spdisp = plt.subplot(2,1,2, sharex=spbet)

spbet.plot(tw.s, tw.betx, label='x')
spbet.plot(tw.s, tw.bety, label='y')
spbet.set_ylabel(r'$\beta_{x,y}$ [m]')
spbet.set_ylim(bottom=0)
spbet.legend(loc='best')

spdisp.plot(tw.s, tw.dx, label='x')
spdisp.plot(tw.s, tw.dy, label='y')
spdisp.set_ylabel(r'$D_{x,y}$ [m]')
spdisp.set_xlabel('s [m]')

fig1.subplots_adjust(left=.15, right=.92, hspace=.27)
plt.show()


import xtrack as xt

# Load a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

tw0 = line.twiss(method='4d')

# Knob optimizer for horizontal chromaticity
opt = line.match_knob('dqx.b1', knob_value_start=tw0.dqx, knob_value_end=3.0,
            run=False, method='4d',
            vary=xt.VaryList(['ksf.b1', 'ksd.b1'], step=1e-8),
            targets=xt.TargetSet(dqx=3.0, dqy=tw0, tol=1e-6))

# New terms have been added to knobs to vary
line.vars['ksf.b1']._expr # is: (0.0 + vars['ksf.b1_from_dqx.b1'])
line.vars['ksd.b1']._expr # is: (0.0 + vars['ksd.b1_from_dqx.b1'])
line.vars['ksf.b1_from_dqx.b1']._expr # is None
line.vars['ksd.b1_from_dqx.b1']._expr # is None

# optimized acts on newly created terms
opt.vary_status(); opt.target_status()
# prints:
#
# Vary status:
# id state tag name               lower_limit current_val upper_limit val_at_iter_0  step weight
#  0 ON        ksf.b1_from_dqx.b1        None           0        None             0 1e-08      1
#  1 ON        ksd.b1_from_dqx.b1        None           0        None             0 1e-08      1
# Target status:
# id state tag tol_met  residue current_val target_val description
#  0 ON          False -1.09005     1.90995          3 'dqx', val=3, tol=1e-06, weight=1
#  1 ON           True        0     1.94297    1.94297 'dqy', val=1.94297, tol=1e-06, weight=1

opt.solve() # perform optimization
opt.vary_status(); opt.target_status()
# prints:
#
# Vary status:
# id state tag name               lower_limit current_val upper_limit val_at_iter_0  step weight
#  0 ON        ksf.b1_from_dqx.b1        None  0.00130336        None             0 1e-08      1
#  1 ON        ksd.b1_from_dqx.b1        None  -0.0004024        None             0 1e-08      1
# Target status:
# id state tag tol_met     residue current_val target_val description
#  0 ON           True 7.94672e-08           3          3 'dqx', val=3, tol=1e-06, weight=1
#  1 ON           True           0     1.94297    1.94297 'dqy', val=1.94297, tol=1e-06, weight=1

# Generate the knob
opt.generate_knob()

line.vars['ksf.b1']._expr # is: (0.0 + vars['ksf.b1_from_dqx.b1'])
line.vars['ksd.b1']._expr # is: (0.0 + vars['ksd.b1_from_dqx.b1'])
line.vars['ksf.b1_from_dqx.b1']._expr
# is ((0.0011956933485755728 * vars['dqx.b1']) - 0.0022837181704350494)
line.vars['ksd.b1_from_dqx.b1']._expr # is None
# is ((-0.0003691583859286993 * vars['dqx.b1']) - -0.0007050751889840094)

# Create also vertical chromaticity knob
opt_dqy = line.match_knob('dqy.b1', knob_value_start=tw0.dqy, knob_value_end=3.0,
            run=False, method='4d',
            vary=xt.VaryList(['ksf.b1', 'ksd.b1'], step=1e-8),
            targets=xt.TargetSet(dqx=tw0, dqy=3.0, tol=1e-6))
opt_dqy.solve()
opt_dqy.generate_knob()

line.vars['ksf.b1']._expr
# is: ((0.0 + vars['ksf.b1_from_dqx.b1']) + vars['ksf.b1_from_dqy.b1'])
line.vars['ksd.b1']._expr
# is: ((0.0 + vars['ksd.b1_from_dqx.b1']) + vars['ksd.b1_from_dqy.b1'])
line.vars['ksf.b1_from_dqx.b1']._expr
# is ((0.0011956933485755728 * vars['dqx.b1']) - 0.0022837181704350494)
line.vars['ksd.b1_from_dqx.b1']._expr # is None
# is ((-0.0003691583859286993 * vars['dqx.b1']) - -0.0007050751889840094)
line.vars['ksf.b1_from_dqy.b1']._expr
# is ((0.0011956933485755728 * vars['dqy.b1']) - 0.0022837181704350494)
line.vars['ksd.b1_from_dqy.b1']._expr
# is ((-0.0003691583859286993 * vars['dqy.b1']) - -0.0007050751889840094)

# Test knobs
line.vars['dqx.b1'] = 5.
line.vars['dqy.b1'] = 6.

tw = line.twiss(method='4d')
tw.dqx # is 5.00000231
tw.dqy # is 5.99999987



import numpy as np

import xobjects as xo
import xtrack as xt

source = """

/*gpukern*/
void get_values_at_offsets_float64(
    MultiSetterData data,
    /*gpuglmem*/ int8_t* buffer,
    /*gpuglmem*/ double* out){

    int64_t num_offsets = MultiSetterData_len_offsets(data);

    for (int64_t ii = 0; ii < num_offsets; ii++) { //vectorize_over ii num_offsets
        int64_t offs = MultiSetterData_get_offsets(data, ii);

        double val = *((/*gpuglmem*/ double*)(buffer + offs));
        out[ii] = val;
    } //end_vectorize
}

/*gpukern*/
void get_values_at_offsets_int64(
    MultiSetterData data,
    /*gpuglmem*/ int8_t* buffer,
    /*gpuglmem*/ int64_t* out){

    int64_t num_offsets = MultiSetterData_len_offsets(data);

    for (int64_t ii = 0; ii < num_offsets; ii++) { //vectorize_over ii num_offsets
        int64_t offs = MultiSetterData_get_offsets(data, ii);

        int64_t val = *((/*gpuglmem*/ int64_t*)(buffer + offs));
        out[ii] = val;
    } //end_vectorize
}

/*gpukern*/
void set_values_at_offsets_float64(
    MultiSetterData data,
    /*gpuglmem*/ int8_t* buffer,
    /*gpuglmem*/ double* input){

    int64_t num_offsets = MultiSetterData_len_offsets(data);

    for (int64_t ii = 0; ii < num_offsets; ii++) {  //vectorize_over ii num_offsets
        int64_t offs = MultiSetterData_get_offsets(data, ii);

        double val = input[ii];
        *((/*gpuglmem*/ double*)(buffer + offs)) = val;
    } //end_vectorize
}

/*gpukern*/
void set_values_at_offsets_int64(
    MultiSetterData data,
    /*gpuglmem*/ int8_t* buffer,
    /*gpuglmem*/ int64_t* input){

    int64_t num_offsets = MultiSetterData_len_offsets(data);

    for (int64_t ii = 0; ii < num_offsets; ii++) {  //vectorize_over ii num_offsets
        int64_t offs = MultiSetterData_get_offsets(data, ii);

        int64_t val = input[ii];
        *((/*gpuglmem*/ int64_t*)(buffer + offs)) = val;
    } //end_vectorize
}



"""

class MultiSetter(xo.HybridClass):
    _xofields = {
        'offsets': xo.Int64[:],
    }

    _extra_c_sources = [
        source,
    ]

    _kernels = {
        'get_values_at_offsets_float64': xo.Kernel(
            c_name='get_values_at_offsets_float64',
            args=[
                xo.Arg(xo.ThisClass, name='data'),
                xo.Arg(xo.Int8, pointer=True, name='buffer'),
                xo.Arg(xo.Float64, pointer=True, name='out'),
            ],
        ),
        'get_values_at_offsets_int64': xo.Kernel(
            c_name='get_values_at_offsets_int64',
            args=[
                xo.Arg(xo.ThisClass, name='data'),
                xo.Arg(xo.Int8, pointer=True, name='buffer'),
                xo.Arg(xo.Int64, pointer=True, name='out'),
            ],
        ),
        'set_values_at_offsets_float64': xo.Kernel(
            c_name='set_values_at_offsets_float64',
            args=[
                xo.Arg(xo.ThisClass, name='data'),
                xo.Arg(xo.Int8, pointer=True, name='buffer'),
                xo.Arg(xo.Float64, pointer=True, name='input'),
            ],
        ),
        'set_values_at_offsets_int64': xo.Kernel(
            c_name='set_values_at_offsets_int64',
            args=[
                xo.Arg(xo.ThisClass, name='data'),
                xo.Arg(xo.Int8, pointer=True, name='buffer'),
                xo.Arg(xo.Int64, pointer=True, name='input'),
            ],
        ),
    }

    def __init__(self, line, elements, field, index=None):
        """Create object to efficiently set and get values of a specific field of
        several elements of a line.

        Parameters
        ----------
        line: xtrack.Line
            Line in which the elements are mutated
        elements: list of int or strings
            List of indices or names of the elements to be mutated.
        field: str
            Name of the field to be mutated.
        index: int or None
            If the field is an array, the index of the array to be mutated.
        """

        if isinstance(line, xt.Tracker):
            tracker = line
        else:
            tracker = line.tracker

        if tracker.iscollective:
            tracker = tracker.line._get_non_collective_line().tracker

        context = tracker._context

        tracker_buffer = tracker._buffer
        line = tracker.line

        if len(elements) == 0:
            self._empty = True
            self.xoinitialize(_context=context, offsets=[])
            return

        self._empty = False

        # Get dtype from first element
        el = line[elements[0]]
        if isinstance(field, (list, tuple)):
            inner_obj = el
            inner_name = field[-1]
            for ff in field[:-1]:
                inner_obj = getattr(inner_obj, ff)
        else:
            inner_obj = el
            inner_name = field
        dd = getattr(inner_obj.copy(_context=xo.context_default), inner_name)
        if index is not None:
            dd = dd[index]
        self.dtype = type(dd)
        self.xodtype = {
            np.float64: xo.Float64,
            np.int64: xo.Int64,
        }[self.dtype]

        assert self.dtype in [np.float64, np.int64], (
            'Only float64 and int64 are supported for now')

        assert np.all([line[nn]._buffer is tracker_buffer for nn in elements])
        offsets = [_extract_offset(line[nn], field, index, self.dtype, self.xodtype)
                   for nn in elements]

        self.xoinitialize(_context=context, offsets=offsets)
        self.compile_kernels(only_if_needed=True)

        self.offsets = context.nparray_to_context_array(np.array(offsets))
        self.tracker = tracker
        self._tracker_buffer = tracker_buffer

        self._get_kernel = {
            np.float64: self._context.kernels.get_values_at_offsets_float64,
            np.int64: self._context.kernels.get_values_at_offsets_int64,
        }[self.dtype]

        self._set_kernel = {
            np.float64: self._context.kernels.set_values_at_offsets_float64,
            np.int64: self._context.kernels.set_values_at_offsets_int64,
        }[self.dtype]

    def get_values(self):
        """Get the values of the multisetter fields."""
        if self._empty:
            return self._context.zeros(0, dtype=np.float64)

        out = self._context.zeros(len(self.offsets), dtype=self.dtype)
        self._get_kernel.set_n_threads(len(self.offsets))
        self._get_kernel(data=self, buffer=self._tracker_buffer.buffer, out=out)
        return out

    def set_values(self, values):
        """Set the values of the multisetter fields.

        Parameters
        ----------
        values: np.ndarray
            Array of values to be set.
        """
        if self._empty:
            return

        self._set_kernel.set_n_threads(len(self.offsets))
        self._set_kernel(data=self, buffer=self._tracker_buffer.buffer,
               input=xt.BeamElement._arr2ctx(self, values))

    def compile_kernels(self, only_if_needed=True):
        context = self._buffer.context
        if context.allow_prebuilt_kernels and only_if_needed:
            try:
                from xsuite import (
                    get_suitable_kernel,
                    XSK_PREBUILT_KERNELS_LOCATION,
                )
                kernel_info = get_suitable_kernel({}, ())
            except ImportError:
                kernel_info = None

            if kernel_info:
                module_name, _ = kernel_info
                kernels = context.kernels_from_file(
                    module_name=module_name,
                    containing_dir=XSK_PREBUILT_KERNELS_LOCATION,
                    kernel_descriptions=self._kernels,
                )
                context.kernels.update(kernels)

        super().compile_kernels(only_if_needed=only_if_needed)


def _extract_offset(obj, field_name, index, dtype, xodtype):

    if isinstance(field_name, (list, tuple)):
        inner_obj = obj
        inner_name = field_name[-1]
        for ff in field_name[:-1]:
            inner_obj = getattr(inner_obj, ff)
    else:
        inner_obj = obj
        inner_name = field_name

    if index is None:
        assert isinstance(getattr(inner_obj, inner_name), dtype), (
            "Inconsistent types")
        return inner_obj._xobject._get_offset(inner_name)
    else:
        assert getattr(inner_obj._xobject, inner_name)._itemtype is xodtype, (
            "Inconsistent types")
        return getattr(inner_obj._xobject, inner_name)._get_offset(index)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib
import pickle

import numpy as np
import pytest

import ducktrack as dtk
import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
@pytest.mark.parametrize(
    'fname_line_particles,tolerances_10_turns,test_backtracker',
    [
        ('lhc_no_bb/line_and_particle.json', (1e-9, 3e-11), True),
        ('hllhc_14/line_and_particle.json', (1e-9, 3e-11), False),
        ('sps_w_spacecharge/line_no_spacecharge_and_particle.json',
            (2e-8, 7e-9), False),
    ],
    ids=['lhc_no_bb', 'hllhc_14', 'sps_w_spacecharge'],
)
def test_full_rings(
    test_context,
    fname_line_particles,
    tolerances_10_turns,
    test_backtracker,
    tmp_path
):
    rtol_10turns, atol_10turns = tolerances_10_turns
    fname_line_particles = test_data_folder.joinpath(fname_line_particles)

    #############
    # Load file #
    #############

    if str(fname_line_particles).endswith('.pkl'):
        with open(fname_line_particles, 'rb') as fid:
            input_data = pickle.load(fid)
    elif str(fname_line_particles).endswith('.json'):
        with open(fname_line_particles, 'r') as fid:
            input_data = json.load(fid)

    ##################
    # Get a line #
    ##################

    line = xt.Line.from_dict(input_data['line'])

    ##################
    # Build TrackJob #
    ##################
    print('Build tracker...')
    line.build_tracker(_context=test_context)
    line.reset_s_at_end_turn = False

    assert line._buffer.context is test_context
    assert line[23]._buffer.context is test_context

    ######################
    # Get some particles #
    ######################
    particles = xp.Particles.from_dict(input_data['particle'],
                                       _context=test_context)
    #########
    # Track #
    #########
    print('Track a few turns...')
    n_turns = 10
    line.track(particles, num_turns=n_turns)

    ###########################
    # Check against ducktrack #
    ###########################

    testline = dtk.TestLine.from_dict(input_data['line'])

    print('Check against ...')
    ip_check = 0
    vars_to_check = ['x', 'px', 'y', 'py', 'zeta', 'delta', 's']
    dtk_part = dtk.TestParticles.from_dict(input_data['particle']).copy()
    for _ in range(n_turns):
        testline.track(dtk_part)

    for vv in vars_to_check:
        dtk_value = getattr(dtk_part, vv)[0]
        xt_value = test_context.nparray_from_context_array(
            getattr(particles, vv))[ip_check]
        passed = np.isclose(xt_value, dtk_value,
                            rtol=rtol_10turns, atol=atol_10turns)
        print(f'Varable {vv}:\n'
              f'    dtk:    {dtk_value: .7e}\n'
              f'    xtrack: {xt_value: .7e}\n')
        if not passed:
            raise ValueError('Discrepancy found!')

    #####################
    # Check backtracker #
    #####################

    if test_backtracker:
        print('Testing backtracker')
        line.track(particles, num_turns=n_turns, backtrack=True)

        dtk_part = dtk.TestParticles(**input_data['particle']).copy()

        for vv in vars_to_check:
            dtk_value = getattr(dtk_part, vv)[0]
            xt_value = test_context.nparray_from_context_array(
                getattr(particles, vv))[ip_check]
            passed = np.isclose(xt_value, dtk_value, rtol=rtol_10turns,
                                atol=atol_10turns)
            if not passed and vv == 's':
                passed = np.isclose(xt_value, dtk_value,
                                    rtol=rtol_10turns, atol=1e-8)

            if not passed:
                print(f'Not passend on backtrack for var {vv}!\n'
                      f'    dtk:    {dtk_value: .7e}\n'
                      f'    xtrack: {xt_value: .7e}\n')
                # raise ValueError
                print('Test passed!')

    ######################
    # Check closed orbit #
    ######################

    part_co = line.find_closed_orbit(co_guess=xp.Particles(
        _context=test_context,
        p0c=input_data['particle']['p0c']))

    parttest = part_co.copy()
    for _ in range(10):
        line.track(parttest)
        xo.assert_allclose(parttest._xobject.x[0], part_co._xobject.x[0],
                          rtol=0, atol=1e-11)
        xo.assert_allclose(parttest._xobject.y[0], part_co._xobject.y[0],
                          rtol=0, atol=1e-11)
        xo.assert_allclose(parttest._xobject.zeta[0], part_co._xobject.zeta[0],
                          rtol=0, atol=5e-11)

@for_all_test_contexts
def test_freeze_vars(test_context):
    test_data_folder.joinpath('hllhc_14/line_and_particle.json'),

    fname_line_particles = test_data_folder.joinpath(
        './hllhc_14/line_and_particle.json')

    #############
    # Load file #
    #############

    with open(fname_line_particles, 'r') as fid:
        input_data = json.load(fid)

    ##############
    # Get a line #
    ##############
    line = xt.Line.from_dict(input_data['line'])

    #################
    # Build Tracker #
    #################
    print('Build tracker...')
    freeze_vars = xp.Particles.part_energy_varnames() + ['zeta']
    line.build_tracker(_context=test_context)
    line.reset_s_at_end_turn = False
    line.freeze_vars(freeze_vars)

    ######################
    # Get some particles #
    ######################
    particle_ref = xp.Particles.from_dict(input_data['particle'])
    particles = xp.build_particles(_context=test_context,
                                   particle_ref=particle_ref,
                                   x=np.linspace(-1e-4, 1e-4, 10))

    particles_before_tracking = particles.copy()

    #########
    # Track #
    #########
    print('Track a few turns...')
    n_turns = 10
    line.track(particles, num_turns=n_turns)

    for vv in ['ptau', 'delta', 'rpp', 'rvv', 'zeta']:
        vv_before = test_context.nparray_from_context_array(
            getattr(particles_before_tracking, vv))
        vv_after = test_context.nparray_from_context_array(
            getattr(particles, vv))
        assert np.all(vv_before == vv_after)

    print('Check passed')
